"""Routines for generating structured meshes on poloidal domains

Classes
-------
`RectangularPoloidalGrid`
    Simple rectangles in R-Z
`StructuredPoloidalGrid`
    Curvilinear structured grids in R-Z

Functions
---------
`grid_annulus`
    Create a `StructuredPoloidalGrid` from inner and outer RZLine
    objects using a simple algorithm
`grid_elliptic`
    Create a `StructuredPoloidalGrid` from inner and outer RZLine
    objects using elliptic meshing method

"""

import warnings

import numpy as np
from numpy import linspace, pi, zeros
from scipy.interpolate import RectBivariateSpline
from scipy.spatial import cKDTree as KDTree

try:
    import matplotlib.pyplot as plt

    plotting_available = True
except ImportError:
    warnings.warn("Couldn't import matplotlib, plotting not available.")
    plotting_available = False

from . import rzline


class PoloidalGrid(object):
    """Represents a poloidal grid

    Note: Here the 2D plane (R,Z) is labelled by (x,z) indices

    Attributes
    ----------
    nx, nz : int
        Number of points in x and z
    R : ndarray
        2D Numpy array of R coordinates
    Z : ndarray
        2D Numpy array of Z coordinates

    """

    def plot(self, axis=None, show=True):
        """Plot grid using matplotlib

        Parameters
        ----------
        axis : matplotlib axis, optional
            A matplotlib axis to plot on. By default a new figure
            is created
        show : bool, optional
            Calls plt.show() at the end

        Returns
        -------
        axis
            The matplotlib axis that was used

        """

        if not plotting_available:
            warnings.warn("matplotlib not available, unable to plot")
            return None

        if axis is None:
            fig = plt.figure()
            axis = fig.add_subplot(1, 1, 1)

        axis.plot(self.R, self.Z, "k-")
        axis.plot(self.R.T, self.Z.T, "k-")

        if show:
            plt.show()

        return axis


class RectangularPoloidalGrid(PoloidalGrid):
    """Represents a poloidal grid consisting of a rectangular domain

    Note: Here the 2D plane (R,Z) is labelled by (x,z) indices

    Attributes
    ----------
    nx, nz : int
        Number of points in x and z
    R : ndarray
        2D Numpy array of R coordinates
    Z : ndarray
        2D Numpy array of Z coordinates

    Parameters
    ----------
    nx : int
        Number of points in major radius (including boundaries)
    nz : int
        Number of points in height (including boundaries)
    Lx : float
        Radial domain size  [m]
    Lz : float
        Vertical domain size [m]
    Rcentre : float, optional
        Coordinate at the middle of the domain
    Zcentre : float, optional
        Coordinate at the middle of the domain
    MXG : int, optional
        Number of guard cells in X. The boundary is put half-way
        between the guard cell and the domain

    """

    def __init__(self, nx, nz, Lx, Lz, Rcentre=0.0, Zcentre=0.0, MXG=2):
        self.nx = nx
        self.nz = nz

        self.Lx = Lx
        self.Lz = Lz

        self.Rcentre = Rcentre
        self.Zcentre = Zcentre

        # Some useful derived quantities
        # Note: index at the middle of the domain is (nx - 1)/2
        #       e.g. nx=5 :  0 1 | 2 | 3 4
        self.dR = self.Lx / (self.nx - 2 * MXG)
        self.dZ = self.Lz / self.nz
        self.Rmin = self.Rcentre - self.dR * (self.nx - 1.0) / 2.0
        self.Zmin = self.Zcentre - self.dZ * (self.nz - 1.0) / 2.0

        # Generate 2D arrays
        # Using getCoordinate to ensure consistency
        xind, zind = np.meshgrid(np.arange(nx), np.arange(nz), indexing="ij")
        self.R, self.Z = self.getCoordinate(xind, zind)

    def __repr__(self):
        return (
            "RectangularPoloidalGrid({0},{1},{2},{3},Rcentre={4},Zcentre={5})".format(
                self.nx, self.nz, self.Lx, self.Lz, self.Rcentre, self.Zcentre
            )
        )

    def getCoordinate(self, xind, zind, dx=0, dz=0):
        """Get coordinates (R,Z) at given (xind,zind) index

        Parameters
        ----------
        xind, zind : array_like
            Indices in X and Z. These should be the same shape
        dx : int, optional
            Order of x derivative
        dz : int, optional
            Order of z derivative

        Returns
        -------
        R, Z : (ndarray, ndarray)
            Locations of point or derivatives of R,Z with respect to
            indices if dx,dz != 0

        """

        # Convert to NumPy arrays if not already
        xind = np.asarray(xind)
        zind = np.asarray(zind)
        # Make sure dx and dz are integers
        dx = int(dx)
        dz = int(dz)

        assert xind.shape == zind.shape
        assert dx >= 0
        assert dz >= 0

        shape = xind.shape

        if dx + dz > 2:
            # Second derivatives and mixed derivatives all zero
            return np.zeros(shape), np.zeros(shape)

        if dx == 1:
            # dR/dx, dZ/dx
            return np.full(shape, self.dR), np.zeros(shape)
        elif dz == 1:
            # dR/dz, dZ/dz
            return np.zeros(shape), np.full(shape, self.dZ)
        # Return (R,Z) location
        return self.Rmin + xind * self.dR, self.Zmin + zind * self.dZ

    def findIndex(self, R, Z):
        """Finds the (x,z) index corresponding to the given (R,Z) coordinate

        Parameters
        ----------
        R, Z : array_like
            Locations to find indices for

        Returns
        -------
        x, z : (ndarray, ndarray)
            Index as a float, same shape as R,Z

        """

        # Make sure inputs are NumPy arrays
        R = np.asarray(R)
        Z = np.asarray(Z)

        # Check that they have the same shape
        assert R.shape == Z.shape

        xind = (R - self.Rmin) / self.dR
        zind = (Z - self.Zmin) / self.dZ

        # Note: These indices may be outside the domain,
        # but this is handled in BOUT++, and useful for periodic
        # domains.

        return xind, zind

    def metric(self):
        """Return the metric tensor, dx and dz

        For this rectangular grid the metric is the identity

        Returns
        -------
        dict
            Dictionary containing:
            - **dx, dz**: Grid spacing
            - **gxx, gxz, gzz**: Covariant components
            - **g_xx, g_xz, g_zz**: Contravariant components
        """
        return {
            "dx": self.dR,
            "dz": self.dZ,  # Grid spacing
            "gxx": 1.0,
            "g_xx": 1.0,
            "gxz": 0.0,
            "g_xz": 0.0,
            "gzz": 1.0,
            "g_zz": 1.0,
        }


class StructuredPoloidalGrid(PoloidalGrid):
    """Represents a structured poloidal grid in R-Z

    Attributes
    ----------
    nx, nz : int
        Number of points in x and z
    R : ndarray
        2D Numpy array of R coordinates
    Z : ndarray
        2D Numpy array of Z coordinates

    Parameters
    ----------
    R, Z : ndarray
        2D Numpy arrays of R,Z points

        .. note:: R,Z are not copied, so these arrays should not be
                  modified afterwards

    """

    def __init__(self, R, Z):
        assert R.shape == Z.shape

        self.R = R
        self.Z = Z

        # Create a KDTree for quick lookup of nearest points
        n = R.size
        data = np.concatenate((R.reshape((n, 1)), Z.reshape((n, 1))), axis=1)
        self.tree = KDTree(data)

        # Create splines for quick interpolation of coordinates
        nx, nz = R.shape

        self.nx = nx
        self.nz = nz

        xinds = np.arange(nx)
        zinds = np.arange(nz * 3)
        # Repeat the data in z, to approximate periodicity
        R_ext = np.concatenate((R, R, R), axis=1)
        Z_ext = np.concatenate((Z, Z, Z), axis=1)

        self._spl_r = RectBivariateSpline(xinds, zinds, R_ext)
        self._spl_z = RectBivariateSpline(xinds, zinds, Z_ext)

    def __repr__(self):
        return "StructuredPoloidalGrid()"

    def getCoordinate(self, xind, zind, dx=0, dz=0):
        """Get coordinates (R, Z) at given (xind, zind) index

        Parameters
        ----------
        xind, zind : array_like
            Indices in X and Z. These should be the same shape
        dx : int, optional
            Order of x derivative
        dz : int, optional
            Order of z derivative

        Returns
        -------
        R, Z : (ndarray, ndarray)
            Locations of point or derivatives of R,Z with respect to
            indices if dx,dz != 0

        """
        nx, nz = self.R.shape
        if (np.amin(xind) < 0) or (np.amax(xind) > nx - 1):
            raise ValueError("x index out of range")

        # Periodic in y
        zind = np.remainder(zind, nz)

        R = self._spl_r(xind, zind + self.nz, dx=dx, dy=dz, grid=False)
        Z = self._spl_z(xind, zind + self.nz, dx=dx, dy=dz, grid=False)

        return R, Z

    def findIndex(self, R, Z, tol=1e-10, show=False):
        """Finds the (x, z) index corresponding to the given (R, Z) coordinate

        Parameters
        ----------
        R, Z : array_like
            Locations. Can be scalar or array, must be the same shape
        tol : float, optional
            Maximum tolerance on the square distance

        Returns
        -------
        x, z : (ndarray, ndarray)
            Index as a float, same shape as R, Z

        """

        # Make sure inputs are NumPy arrays
        R = np.asarray(R)
        Z = np.asarray(Z)

        # Check that they have the same shape
        assert R.shape == Z.shape

        input_shape = R.shape  # So output has same shape as input

        # Get distance and index into flattened data
        # Note ind can be an integer, or an array of ints
        # with the same number of elements as the input (R,Z) arrays
        n = R.size
        position = np.concatenate((R.reshape((n, 1)), Z.reshape((n, 1))), axis=1)

        R = R.reshape((n,))
        Z = Z.reshape((n,))

        dists, ind = self.tree.query(position)

        # Calculate (x,y) index
        nx, nz = self.R.shape
        xind = np.floor_divide(ind, nz)
        zind = ind - xind * nz

        # Convert indices to float
        xind = np.asarray(xind, dtype=float)
        zind = np.asarray(zind, dtype=float)

        # Create a mask for the positions
        mask = np.ones(xind.shape)
        mask[np.logical_or((xind < 0.5), (xind > (nx - 1.5)))] = (
            0.0  # Set to zero if near the boundary
        )

        if show and plotting_available:
            plt.plot(self.R, self.Z, ".")
            plt.plot(R, Z, "x")

        cnt = 0
        underrelax = 1

        while True:
            # Use Newton iteration to find the index
            # dR, dZ are the distance away from the desired point
            Rpos, Zpos = self.getCoordinate(xind, zind)
            if show and plotting_available:
                plt.plot(Rpos, Zpos, "o")
            dR = Rpos - R
            dZ = Zpos - Z

            # Check if close enough
            # Note: only check the points which are not in the boundary
            val = np.amax(mask * (dR**2 + dZ**2))
            if val < tol:
                break
            cnt += 1
            if cnt == 10:
                underrelax = 1.5
            if cnt == 100:
                underrelax = 2
            if cnt == 300:
                underrelax = 2.5
            if cnt == 700:
                underrelax = 3
            if cnt == 1000:
                raise RuntimeError("Failed to converge")

            # Calculate derivatives
            dRdx, dZdx = self.getCoordinate(xind, zind, dx=1)
            dRdz, dZdz = self.getCoordinate(xind, zind, dz=1)

            # Invert 2x2 matrix to get change in coordinates
            #
            # (x) -=  ( dR/dx   dR/dz )^-1  (dR)
            # (y)     ( dZ/dx   dZ/dz )     (dz)
            #
            #
            # (x) -=  ( dZ/dz  -dR/dz ) (dR)
            # (y)     (-dZ/dx   dR/dx ) (dZ) / (dR/dx*dZ/dy - dR/dy*dZ/dx)
            determinant = dRdx * dZdz - dRdz * dZdx

            xind -= mask * ((dZdz * dR - dRdz * dZ) / determinant / underrelax)
            zind -= mask * ((dRdx * dZ - dZdx * dR) / determinant / underrelax)

            # Re-check for boundary
            in_boundary = xind < 0.5
            mask[in_boundary] = 0.0  # Set to zero if near the boundary
            xind[in_boundary] = 0.0
            out_boundary = xind > (nx - 1.5)
            mask[out_boundary] = 0.0  # Set to zero if near the boundary
            xind[out_boundary] = nx - 1

        if show and plotting_available:
            plt.show()

        # Set xind to -1 if in the inner boundary, nx if in outer boundary
        in_boundary = xind < 0.5
        xind[in_boundary] = -1
        out_boundary = xind > (nx - 1.5)
        xind[out_boundary] = nx

        return xind.reshape(input_shape), zind.reshape(input_shape)

    def metric(self):
        """Return the metric tensor, dx and dz

        Returns
        -------
        dict
            Dictionary containing:
            - **dx, dz**: Grid spacing
            - **gxx, gxz, gzz**: Covariant components
            - **g_xx, g_xz, g_zz**: Contravariant components

        """

        # Get arrays of indices
        xind, zind = np.meshgrid(np.arange(self.nx), np.arange(self.nz), indexing="ij")

        # Calculate the gradient along each coordinate
        dolddnew = np.array(
            [self.getCoordinate(xind, zind, dx=a, dz=b) for a, b in ((1, 0), (0, 1))]
        )
        # dims: 0 : dx or dz?
        #       1 : R or z?
        #       2 : spatial: r
        #       3 : spatial: \theta
        ddist = np.sqrt(np.sum(dolddnew**2, axis=1))  # sum R + z
        nx, nz = ddist.shape[1:]
        ddist[0] = 1 / nx
        ddist[1] = 1 / nz
        dolddnew /= ddist[:, None, ...]

        # g_ij = J_ki J_kj
        # (2.5.27) from D'Haeseleer 1991
        # Note: our J is transposed
        J = dolddnew
        g = np.sum(
            np.array(
                [
                    [[J[j, i] * J[k, i] for i in range(2)] for j in range(2)]
                    for k in range(2)
                ]
            ),
            axis=2,
        )

        assert np.all(
            g[0, 0] > 0
        ), f"g[0, 0] is expected to be positive, but some values are not (minimum {np.min(g[0, 0])})"
        assert np.all(
            g[1, 1] > 0
        ), f"g[1, 1] is expected to be positive, but some values are not (minimum {np.min(g[1, 1])})"
        g = g.transpose(2, 3, 0, 1)
        assert np.all(
            np.linalg.det(g) > 0
        ), f"All determinants of g should be positive, but some are not (minimum {np.min(np.linalg.det(g))})"
        ginv = np.linalg.inv(g)
        # Jacobian from BOUT++
        JB = self.R * (J[0, 0] * J[1, 1] - J[0, 1] * J[1, 0])
        return {
            "dx": ddist[0],
            "dz": ddist[1],  # Grid spacing
            "gxx": ginv[..., 0, 0],
            "g_xx": g[..., 0, 0],
            "gxz": ginv[..., 0, 1],
            "g_xz": g[..., 0, 1],
            "gzz": ginv[..., 1, 1],
            "g_zz": g[..., 1, 1],
            # "J": JB,
        }


def grid_annulus(inner, outer, nx, nz, show=True, return_coords=False):
    """Grid an annular region, given inner and outer boundaries both of
    which are RZline objects

    This is a very simple algorithm which just draws straight lines
    between inner and outer boundaries.

    Parameters
    ----------
    inner, outer : `RZline`
        Inner and outer boundaries of the domain
    nx : int
        The required radial resolution, including boundaries
    nz : int
        The required poloidal resolution
    show : bool, optional
        If True, plot the resulting grid
    return_coords : bool, optional
        If True, return the R, Z coordinates of the grid points,
        instead of a `StructuredPoloidalGrid`

    Returns
    -------
    StructuredPoloidalGrid
        A grid of the region

    """

    assert nx >= 2
    assert nz > 1

    R = zeros((nx, nz))
    Z = zeros((nx, nz))

    # Generate angle values, which should now be equally spaced
    # in distance along inner and outer boundaries
    thetavals = linspace(0, 2 * pi, nz, endpoint=False)

    # Radial coordinate
    xvals = linspace(0, 1.0, nx, endpoint=True)

    innerR = inner.Rvalue(thetavals)
    innerZ = inner.Zvalue(thetavals)

    outerR = outer.Rvalue(thetavals)
    outerZ = outer.Zvalue(thetavals)
    for i, x in enumerate(xvals):
        # Get the R and Z coordinates of this line
        R[i, :] = x * outerR + (1.0 - x) * innerR
        Z[i, :] = x * outerZ + (1.0 - x) * innerZ

    if show and plotting_available:
        plt.plot(inner.R, inner.Z, "-o")
        plt.plot(outer.R, outer.Z, "-o")

        plt.plot(R, Z, "x")

        plt.show()

    if return_coords:
        return R, Z
    return StructuredPoloidalGrid(R, Z)


def grid_elliptic(
    inner,
    outer,
    nx,
    nz,
    show=False,
    tol=1e-10,
    align=True,
    restrict_size=20,
    restrict_factor=2,
    return_coords=False,
    nx_outer=0,
    nx_inner=0,
    legacy_align=False,
    inner_ort=True,
    maxfac_inner=None,
    dz_relax=None,
):
    """Create a structured grid between inner and outer boundaries using
    elliptic method

    Coordinates x = x(R, Z) and z = z(R,Z) obey an elliptic equation:

    .. math::

        d^2x/dR^2 + d^2x/dZ^2 = 0

        d^2z/dR^2 + d^2z/dZ^2 = 0

    where here x is in in the domain (0, 1) and z in (0, 2pi)

    The above equations are inverted, giving:

    .. math::

        a*R_xx - 2*b*R_xz + c*R_zz = 0

        a*Z_xx - 2*b*Z_xz + c*Z_zz = 0

    where

    .. math::

        a &= R_z^2 + Z_z^2

        b &= R_z*R_x + Z_x*Z_z

        c &= R_x^2 + Z_x^2

    This is a nonlinear system of equations which is solved
    iteratively.

    Parameters
    ----------
    inner, outer : `RZline`
        Inner and outer boundaries of the domain
    nx : int
        The required radial resolution, including boundaries
    nz : int
        The required poloidal resolution
    show : bool, optional
        Display plots of intermediate results
    tol : float, optional
        Controls when iteration stops
    align : bool, optional
        Attempt to align the inner and outer boundaries
    restrict_size : int, optional
        The size (nx or nz) above which the grid is coarsened
    restrict_factor : int, optional
        The factor by which the grid is divided if coarsened
    nx_outer: int, optional
        The number of additional points outside of the outer boundary
    nx_inner: int, optional
        The number of additional points outside of the inner boundary
    inner_ort: bool, optional
        Whether to place the inner points as close as possible to the
        corresponding outer ones. That increases orthogonality of the grid.
    maxfac_inner : int, optional
        If given, ensure the spacing between smalles and largest distance on
        the inner boundary. Only used with inner_ort. Must be larger than 1.
    return_coords : bool, optional
        If True, return the R, Z coordinates of the grid points,
        instead of a `StructuredPoloidalGrid`

    Returns
    -------
    If return_coords is true, returns R,Z as arrays.
    If return_coords is false, returns a `StructuredPoloidalGrid` object

    References
    ----------
    https://web.archive.org/web/20040530002835/https://www.nada.kth.se/kurser/kth/2D1263/l2.pdf
    https://en.wikipedia.org/wiki/Principles_of_grid_generation

    """
    if nx_outer:
        assert nx_outer > 0
        nx -= nx_outer
    if nx_inner:
        assert nx_inner > 0
        nx -= nx_inner

    assert nx > 1
    assert nz > 1

    # Generate angle values (y coordinate),
    # which should now be equally spaced
    # in distance along inner and outer boundaries
    thetavals = linspace(0, 2 * pi, nz, endpoint=False)

    # Radial coordinate
    xvals = linspace(0, 1.0, nx, endpoint=True)

    if align and not inner_ort:
        # Align inner and outer boundaries
        # Easiest way is to roll both boundaries
        # so that index 0 is on the outboard midplane.
        # This is useless if inner_ort is used, where the inner points are
        # based on a project of the outer points.
        if len(inner.R) < len(outer.R):
            shorter = inner
            longer = outer
        else:
            shorter = outer
            longer = inner
        ind = np.argmax(shorter.R)
        shorter = rzline.RZline(np.roll(shorter.R, -ind), np.roll(shorter.Z, -ind))
        if legacy_align:
            ind = np.argmax(longer.R)
            longer = rzline.RZline(np.roll(longer.R, -ind), np.roll(longer.Z, -ind))
        else:
            dr = shorter.R - longer.R[:, None]
            dz = shorter.Z - longer.Z[:, None]
            delta = dr**2 + dz**2
            fac = len(longer.R) / len(shorter.R)
            j = np.arange(len(shorter.R), dtype=int)
            sums = [
                np.sum(delta[np.round(j * fac).astype(int) - i, j])
                for i in range(len(longer.R))
            ]
            ind = -np.argmin(sums)
            longer = rzline.RZline(np.roll(longer.R, -ind), np.roll(longer.Z, -ind))
        if len(inner.R) < len(outer.R):
            inner = shorter
            outer = longer
        else:
            outer = shorter
            inner = longer

    def laplace(x):
        def dx(x, sign):
            dxv = np.roll(x, sign) - x
            dxv = np.remainder(dxv + np.pi, 2 * np.pi) - np.pi
            return dxv

        fac = 0.1 * nz / 192
        steps = 10
        if fac > 0.1:
            steps = int(steps * fac / 0.1)
            fac = 0.1
        x0 = x.copy()
        for i in range(steps):
            x += fac * (dx(x, 1) + dx(x, -1))

        extra = 0
        if maxfac_inner:

            def getfac(x):
                d = dx(x, -1)
                return np.max(d) / np.min(d)

            while not np.all(dx(x, -1) > 0):
                x += fac * (dx(x, 1) + dx(x, -1))
                extra += 1

            if maxfac_inner == 1:
                x1 = x.copy()
                for i in range(1, len(x0)):
                    if x1[i - 1] > x1[i]:
                        x1[i] += 2 * np.pi
                x2 = (
                    np.mean(x1)
                    + np.linspace(-np.pi, np.pi, len(x0), endpoint=False)
                    + np.pi / len(x0)
                )
                gf = getfac(x0)
                ratio = (maxfac_inner - 1) / (gf - 1)
                x = x2 * (1 - ratio) + np.array(x1) * (ratio)
                x %= 2 * np.pi
            else:
                while getfac(x) > maxfac_inner:
                    extra += 1
                    x += fac * (dx(x, 1) + dx(x, -1))
        else:
            while not np.all(dx(x, -1) > 0):
                x += fac * (dx(x, 1) + dx(x, -1))
                extra += 1
            if extra:
                print(f"Required {extra} extra steps in smoothing!")

        if not np.all(dx(x, -1) > 0):
            plt.plot(x, label="result")
            plt.plot(x0, label="init")
            plt.plot(dx(x0, -1))
            plt.plot(dx(x, -1))
            plt.legend()
            plt.title("Not monotonic!")
            plt.show()
            assert False
        return x

    if (nx > restrict_size) or (nz > restrict_size):
        # Create a coarse grid first to get a starting guess
        # Only restrict the dimensions which exceed restrict_size
        # Note that this might result in multiple levels of resolution

        nx_r = nx
        if nx > restrict_size:
            nx_r = int(nx / restrict_factor)

        nz_r = nz
        if nz > restrict_size:
            nz_r = int(nz / restrict_factor)

        # Create the coarse mesh
        R_r, Z_r = grid_elliptic(
            inner,
            outer,
            nx_r,
            nz_r,
            align=False,
            tol=tol,
            restrict_size=restrict_size,
            restrict_factor=restrict_factor,
            return_coords=True,
            inner_ort=inner_ort,
            maxfac_inner=maxfac_inner,
            dz_relax=dz_relax,
        )

        # Note: Lower case x,z are indices
        z_r = linspace(
            0, 2 * pi, nz_r + 1, endpoint=True
        )  # Add on the final point duplicating the first
        x_r = linspace(0, 1.0, nx_r, endpoint=True)

        # Note: Upper case R,Z are real-space locations
        R_r = np.concatenate((R_r, np.reshape(R_r[:, 0], (nx_r, 1))), axis=1)
        Z_r = np.concatenate((Z_r, np.reshape(Z_r[:, 0], (nx_r, 1))), axis=1)

        # Now interpolate
        spl = RectBivariateSpline(x_r, z_r, R_r)
        R = spl(xvals, thetavals, grid=True)
        spl = RectBivariateSpline(x_r, z_r, Z_r)
        Z = spl(xvals, thetavals, grid=True)

        # Make sure that the inner and outer boundaries are on the
        # inner and outer RZline, not interpolated
        R[-1, :] = outer.Rvalue(thetavals)
        Z[-1, :] = outer.Zvalue(thetavals)

        if inner_ort:
            thetavals_inner = [inner.closestPoint(*x) for x in zip(R[-1], Z[-1])]
            thetavals_inner = laplace(thetavals_inner)

        else:
            thetavals_inner = thetavals

        R[0, :] = inner.Rvalue(thetavals_inner)
        Z[0, :] = inner.Zvalue(thetavals_inner)

    else:
        # Interpolate coordinates of inner and outer boundary
        Router = outer.Rvalue(thetavals)
        Zouter = outer.Zvalue(thetavals)

        if inner_ort:
            thetavals_inner = [inner.closestPoint(*x) for x in zip(Router, Zouter)]
            thetavals_inner = laplace(thetavals_inner)
        else:
            thetavals_inner = thetavals

        Rinner = inner.Rvalue(thetavals_inner)
        Zinner = inner.Zvalue(thetavals_inner)

        # Interpolate in x between inner and outer
        # to get starting guess for a grid
        R = xvals[:, None] * Router[None, :] + (1 - xvals[:, None]) * Rinner[None, :]
        Z = xvals[:, None] * Zouter[None, :] + (1 - xvals[:, None]) * Zinner[None, :]

    dx = xvals[1] - xvals[0]
    dz = thetavals[1] - thetavals[0]

    if show and plotting_available:
        # Markers on original points on inner and outer boundaries
        plt.plot(inner.R, inner.Z, "-o")
        plt.plot(outer.R, outer.Z, "-o")

        # Black lines through inner and outer boundaries
        r, z = inner.position(np.linspace(0, 2 * np.pi, 10 * nz))
        plt.plot(r, z, "k")
        r, z = outer.position(np.linspace(0, 2 * np.pi, 10 * nz))
        plt.plot(r, z, "k")

        # Red dots to mark the inner and outer boundaries
        plt.plot(R[0, :], Z[0, :], "ro")
        plt.plot(R[-1, :], Z[-1, :], "ro")

    # Start solver loop
    for _ in range(R.size * 2):
        # Calculate coefficients, which exclude boundary points
        # Note that the domain is periodic in y so roll arrays

        R_xm = R[:-2, :]  # R(x-1,z)
        R_xp = R[2:, :]  # R(x+1,z)
        R_zm = np.roll(R, 1, axis=1)  # R(x, z-1)
        R_zp = np.roll(R, -1, axis=1)  # R(x, z+1)
        R_xmzm = R_zm[:-2, :]  # R(x-1, z-1)
        R_xpzm = R_zm[2:, :]  # R(x+1, z-1)
        R_xmzp = R_zp[:-2, :]  # R(x-1, z+1)
        R_xpzp = R_zp[2:, :]  # R(x+1, z+1)
        R_zm = R_zm[1:-1, :]  # Now chop off x boundaries
        R_zp = R_zp[1:-1, :]  # This is done to minimise number of rolls

        Z_xm = Z[:-2, :]
        Z_xp = Z[2:, :]
        Z_zm = np.roll(Z, 1, axis=1)
        Z_zp = np.roll(Z, -1, axis=1)
        Z_xmzm = Z_zm[:-2, :]
        Z_xpzm = Z_zm[2:, :]
        Z_xmzp = Z_zp[:-2, :]
        Z_xpzp = Z_zp[2:, :]
        Z_zm = Z_zm[1:-1, :]
        Z_zp = Z_zp[1:-1, :]

        dz_relax = dz_relax or 10 / 3
        dRdz = dz_relax * (R_zp - R_zm) / (2.0 * dz)
        dRdx = (R_xp - R_xm) / (2.0 * dx)

        dZdz = dz_relax * (Z_zp - Z_zm) / (2.0 * dz)
        dZdx = (Z_xp - Z_xm) / (2.0 * dx)

        a = dRdz**2 + dZdz**2
        b = dRdz * dRdx + dZdx * dZdz
        c = dRdx**2 + dZdx**2

        # Now solve a*R_xx - 2*b*R_xz + c*R_zz = 0
        # For now using Jacobi update

        a_dx2 = a / dx**2
        b_dxdz = b / (2.0 * dx * dz)
        c_dz2 = c / dz**2
        inv_diag = 1.0 / (2 * a / dx**2 + 2 * c / dz**2)

        Rold = R.copy()
        Zold = Z.copy()

        R[1:-1, :] = (
            a_dx2 * (R_xm + R_xp)
            - b_dxdz * (R_xpzp - R_xmzp - R_xpzm + R_xmzm)
            + c_dz2 * (R_zm + R_zp)
        ) * inv_diag

        Z[1:-1, :] = (
            a_dx2 * (Z_xm + Z_xp)
            - b_dxdz * (Z_xpzp - Z_xmzp - Z_xpzm + Z_xmzm)
            + c_dz2 * (Z_zm + Z_zp)
        ) * inv_diag

        maxchange_sq = np.amax((R - Rold) ** 2 + (Z - Zold) ** 2)

        if maxchange_sq < tol:
            break

    else:
        print("Convergence failure, trying to plot ...")
        if plotting_available:
            if not show:
                # Markers on original points on inner and outer boundaries
                plt.plot(inner.R, inner.Z, "-o", label="inner")
                plt.plot(outer.R, outer.Z, "-o", label="outer")

                # Black lines through inner and outer boundaries
                r, z = inner.position(np.linspace(0, 2 * np.pi, 10 * nz))
                plt.plot(r, z, "k")
                r, z = outer.position(np.linspace(0, 2 * np.pi, 10 * nz))
                plt.plot(r, z, "k")

                # Red dots to mark the inner and outer boundaries
                plt.plot(R[0, :], Z[0, :], "ro")
                plt.plot(R[-1, :], Z[-1, :], "ro")
            plt.show()

        raise RuntimeError(
            f"Failed to converge - grid shape {R.shape}: Error was {maxchange_sq} > {tol}"
        )

    if nx_outer:
        dn = np.arange(nx_outer) + 1
        dR = R[-1] - R[-2]
        dZ = Z[-1] - Z[-2]
        R = np.vstack([R, R[-1] + dR[None, :] * dn[:, None]])
        Z = np.vstack([Z, Z[-1] + dZ[None, :] * dn[:, None]])
    if nx_inner:
        dn = np.arange(nx_inner)[::-1] + 1
        dR = R[0] - R[1]
        dZ = Z[0] - Z[1]
        R = np.vstack([R[0] + dR[None, :] * dn[:, None], R])
        Z = np.vstack([Z[0] + dZ[None, :] * dn[:, None], Z])

    if show and plotting_available:
        plt.plot(R, Z)
        plt.plot(np.transpose(R), np.transpose(Z))
        plt.show()

    if return_coords:
        return R, Z
    return StructuredPoloidalGrid(R, Z)


if __name__ == "__main__":
    inner = rzline.shaped_line(R0=3.0, a=0.5, elong=1.0, triang=0.0, indent=1.0, n=50)
    outer = rzline.shaped_line(R0=2.8, a=1.5, elong=1.0, triang=0.0, indent=0.2, n=50)

    grid = grid_elliptic(inner, outer, 100, 100, show=True)

    x, z = grid.findIndex([2.0, 1.9], [1.5, 2.0])
    print(x, z)



import argparse

import numpy as np
from boututils.boutarray import BoutArray

import zoidberg as zb

from .common import calc_curvilinear_curvature


def screwpinch(
    nx=68,
    ny=16,
    nz=128,
    xcentre=1.5,
    fname="screwpinch.fci.nc",
    a=0.2,
    npoints=421,
    show_maps=False,
):
    yperiod = 2 * np.pi
    field = zb.field.Screwpinch(xcentre=xcentre, yperiod=yperiod, shear=0)
    ycoords = np.linspace(0.0, yperiod, ny, endpoint=False)
    print("Making curvilinear poloidal grid")
    inner = zb.rzline.shaped_line(
        R0=xcentre, a=a / 2.0, elong=0, triang=0.0, indent=0, n=npoints
    )
    outer = zb.rzline.shaped_line(
        R0=xcentre, a=a, elong=0, triang=0.0, indent=0, n=npoints
    )
    poloidal_grid = zb.poloidal_grid.grid_elliptic(inner, outer, nx, nz)
    grid = zb.grid.Grid(poloidal_grid, ycoords, yperiod, yperiodic=True)
    maps = zb.make_maps(grid, field)
    maps["y_coord"] = BoutArray(ycoords, dict(bout_type="ArrayY"))
    zb.write_maps(grid, field, maps, str(fname), metric2d=False)
    calc_curvilinear_curvature(fname, field, grid, maps)

    if show_maps:
        zb.plot.plot_forward_map(grid, maps, yslice=0)


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument("--nx", "-x", help="number of x points", default=68, type=int)
    parser.add_argument("--ny", "-y", help="number of y points", default=16, type=int)
    parser.add_argument("--nz", "-z", help="number of z points", default=128, type=int)
    parser.add_argument(
        "--xcentre", "-c", help="Centre of the screw pinch", default=1.5, type=float
    )
    parser.add_argument(
        "--fname",
        "-f",
        help="File name of the grid to be written",
        default="screwpinch.fci.nc",
        type=str,
    )
    parser.add_argument("-a", help="???", default=0.2, type=float)
    parser.add_argument(
        "--npoints",
        "-p",
        help="Number of iterations for surface tracing. Needs to be large enough.",
        default=421,
        type=int,
    )

    parser.add_argument("--show_maps", "-s", help="plot the grid", action="store_true")

    args = parser.parse_args()
    screwpinch(**args.__dict__)


if __name__ == "__main__":
    main()



import numpy as np

from . import rzline


def test_circular_boundaries():
    inner = rzline.circle(R0=1.0, r=1.0, n=20)
    outer = rzline.circle(R0=1.0, r=2.0, n=20)

    assert np.amax(inner.R) <= 2.0
    assert np.amin(inner.R) >= 0.0


def test_distance():
    # Check the RZline.distance() function

    n = 50
    inner = rzline.circle(R0=1.0, r=1.0, n=n)
    outer = rzline.circle(R0=1.0, r=2.0, n=n)

    dist = inner.distance()

    # Check size of the array
    assert len(dist) == n + 1

    # Check start and end values
    assert np.allclose(dist[0], 0.0)
    assert np.allclose(dist[-1], 2.0 * np.pi)

    # Check that dist is monotonically increasing
    assert np.all(np.greater(dist[1:] - dist[:-1], 0.0))

    dist_outer = outer.distance()
    assert np.allclose(dist_outer[-1], 4.0 * np.pi)


def test_order_by_distance():
    # Check the RZline.equallySpaced function
    inner = rzline.circle(R0=1.0, r=1.0, n=20)

    new_inner = inner.equallySpaced(n=10)

    assert len(new_inner.R) == 10


def test_line_from_points():
    # Create a shaped periodic line
    original = rzline.shaped_line(
        R0=3.0, a=1.0, elong=1.0, triang=0.4, indent=1.0, n=20
    )

    # Permute points
    from numpy import random

    random.seed(1235)  # Fix seed so same every time
    indx = random.permutation(original.R.size)

    R = original.R[indx]
    Z = original.Z[indx]

    reconstructed = rzline.line_from_points(R, Z)

    # Reconstructed should now be a rotated version of original
    match = False
    for i in range(len(R)):
        if np.allclose(original.R, np.roll(reconstructed.R, i)) and np.allclose(
            original.Z, np.roll(reconstructed.Z, i)
        ):
            match = True
            break

    assert match



try:
    from importlib.metadata import PackageNotFoundError, version
except (ModuleNotFoundError, ImportError):
    from importlib_metadata import PackageNotFoundError, version
try:
    __version__ = version("zoidberg")
except PackageNotFoundError:
    from setuptools_scm import get_version

    __version__ = get_version(root="..", relative_to=__file__)

from . import field, fieldtracer, grid, plot, zoidberg
from .zoidberg import make_maps, write_maps

__all__ = [
    "__version__",
    "field",
    "fieldtracer",
    "grid",
    "make_maps",
    "plot",
    "write_maps",
    "zoidberg",
]



import numpy as np
import xarray as xr
import matplotlib.pyplot as plt
from mpl_toolkits.axes_grid1 import make_axes_locatable
import boutpp as bc
import boutconfig

assert boutconfig.isMetric3D(), f"""Require 3D metric - but boutpp seems to be compiled with 2D Metrics
NB: boutpp     is {bc.__path__}
    boutconfig is {boutconfig.__path__}"""

lst = [4, 8, 16, 32, 64]


def extend(p):
    # print(p.shape, p.dims)
    out = np.concatenate([p[:, -1:, :], p, p[:, :1, :]], axis=1)
    # print(out.shape)
    return out


def diff(p, axis):
    if axis == 0:
        out = np.zeros_like(p)
        out[:-1] = p[1:] - p[:-1]
        out[1:] += p[1:] - p[:-1]
        out[1:-1] /= 2
        return out
    return (np.roll(p, -1, axis=axis) - np.roll(p, 1, axis=axis)) / 2


def diff2(p, axis):
    assert axis != 0
    return (2 * p - np.roll(p, 1, axis=axis) - np.roll(p, -1, axis=axis)) / 2


def clean(p):
    return p[:, 1:-1, :]


def test(fn, inpf, anaf, testfunc, iname):
    grid = xr.open_dataset(fn)
    bc.Options.root().set("mesh:file", fn, force=True)
    mesh = bc.Mesh(section="")

    f = bc.create3D("0", mesh)

    inp = inpf(grid)
    ana = anaf(grid)

    # Set input field
    f[:, :, :] = inp
    mesh.communicate(f)
    calc = testfunc(f).get()
    l2 = np.sqrt(np.mean(clean(ana - calc)[1:-1] ** 2))

    if 1:
        fig, axs = plt.subplots(1, 4)
        calcd = clean(calc)[:, 1]
        rz = [clean(extend(grid[d]))[:, 1] for d in "RZ"]
        anad = ana[:, 1]
        for dat, label, ax in zip(
            [clean(inp)[:, 1], anad, calcd, (calcd - anad)],
            ["inp", "ana", "calc", "err"],
            axs,
        ):
            plot = ax.pcolormesh(rz[0][1:-1, 1:-1], rz[1][1:-1, 1:-1], dat[1:-1, 1:-1])
            ax.set_title(fn[-20:] + " " + label + " " + iname)
            divider = make_axes_locatable(ax)
            cax = divider.append_axes("right", size="5%", pad=0.05)
            plt.colorbar(plot, cax=cax, orientation="vertical")
    return l2


def do_tests(grids, inpf, anaf, iname, testf, expect=2):
    fail = False
    for mode, todo in grids.items():
        l2 = [test(x[0], inpf, anaf, testf, iname) for x in todo]
        lst = [x[-1] for x in todo]

        errc = np.log(l2[-2] / l2[-1])
        difc = np.log(lst[-1] / lst[-2])
        conv = errc / difc
        info = todo[0][0]
        info = "_".join(info.split("_")[:2])
        print(info, mode, todo, conv, l2)
        with open(f"result_{info}_{iname}.txt", "w") as f:
            f.write(info)
            f.write("\n")
            f.write(iname)
            f.write("\n")
            f.write(" ".join([str(x) for x in lst]))
            f.write("\n")
            f.write(" ".join([str(x) for x in l2]))
            f.write("\n")
        passes = np.isclose(conv, expect, atol=0.3)
        if not passes:
            fail = True
        print(
            f"{conv} {'=' if passes else '!'}= {expect} - {'passing :-)' if passes else 'failing!'}"
        )
        if passes:
            plt.close("all")
        plt.show()
    assert not fail



import warnings

import numpy as np

from . import fieldtracer

try:
    import matplotlib.animation as anim
    import matplotlib.pyplot as plt

    plotting_available = True
except ImportError:
    warnings.warn("Couldn't import matplotlib, plotting not available.")
    plotting_available = False


def plot_poincare(
    magnetic_field,
    xpos,
    zpos,
    yperiod,
    nplot=3,
    y_slices=None,
    revs=40,
    nover=20,
    interactive=False,
):
    """Plot a Poincare graph of the field lines.

    Parameters
    ----------
    magnetic_field : :py:obj:`zoidberg.field.MagneticField`
        Magnetic field object
    xpos, zpos : array_like
        Starting X, Z locations
    yperiod : float
        Length of period in y domain
    nplot : int, optional
        Number of equally spaced y-slices to plot
    y_slices : list of int, optional
        List of y-slices to plot; overrides nplot
    revs : int, optional
        Number of revolutions (times around phi)
    interactive : bool, optional
        If True, plots can be interacted with via the mouse:
        - Left-click on the plot to trace a field-line from that point
        - Right-click to add an additional trace
        - Middle-click to clear added traces

    Returns
    -------
    fig, ax
        The matplotlib figure and axis used
    """

    if not plotting_available:
        warnings.warning("matplotlib not available, unable to plot")
        return

    # Get Poincare plot
    result, y_slices = fieldtracer.trace_poincare(
        magnetic_field,
        xpos,
        zpos,
        yperiod,
        nplot=nplot,
        y_slices=y_slices,
        revs=revs,
        nover=nover,
    )
    nplot = len(y_slices)

    #######################################################
    # Plotting

    colours = ["k", "b", "r", "g", "c", "m"]
    if nplot > len(colours):
        colours += colours * np.floor(nplot / len(colours))

    fig, ax = plt.subplots(1, 1)

    for index in range(nplot):
        r = result[:, index, ..., 0]
        z = result[:, index, ..., 1]
        style = {
            "marker": ".",
            "color": colours[index],
            "linestyle": "None",
        }
        ax.plot(r, z, **style)

    ax.set_xlabel("Radius [m]", fontsize=20)
    ax.set_ylabel("Height [m]", fontsize=20)
    ax.tick_params(axis="both", labelsize=15)

    for phi, colour in zip(y_slices, colours):
        ax.plot([], [], color=colour, label=r"$Y = {0:.2f}$".format(phi))

    ax.legend()

    (overplot,) = ax.plot([], [], "ok")

    def onclick(event):
        # Check if user clicked inside the plot
        if event.xdata is None:
            return

        # Middle mouse
        if event.button == 2:
            x_ = []
            z_ = []
        else:
            result = field_tracer.follow_field_lines(event.xdata, event.ydata, y_values)
            # Right mouse
            if event.button == 3:
                x_, z_ = overplot.get_data()
                x_ = np.append(x_, result[:, 0])
                z_ = np.append(z_, result[:, 1])
            # Left mouse
            else:
                x_ = result[:, 0]
                z_ = result[:, 1]

        overplot.set_data(x_, z_)
        ax.relim()
        ax.autoscale_view()
        fig.canvas.draw()

    if interactive:
        field_tracer = fieldtracer.FieldTracer(magnetic_field)

        revs = int(revs)
        y_values = y_slices[:]
        for n in np.arange(1, revs):
            y_values = np.append(y_values, n * yperiod + y_values[:nplot])

        fig.canvas.mpl_connect("button_press_event", onclick)
    plt.show()

    return fig, ax


def plot_3d_field_line(magnetic_field, xpos, zpos, yperiod, cycles=20, y_res=50):
    """Make a 3D plot of field lines

    Parameters
    ----------
    magnetic_field : :py:obj:`zoidberg.field.MagneticField`
        Magnetic field object
    xpos, zpos : array_like
        Starting X, Z locations
    yperiod : float
        Length of period in y domain
    cycles : int, optional
        Number of times to go round in y
    y_res : int, optional
        Number of points in y in each cycle

    Returns
    -------
    fig, ax
        The matplotlib figure and axis used

    """

    if not plotting_available:
        warnings.warning("matplotlib not available, unable to plot")
        return

    yperiod = float(yperiod)

    # Go round toroidally cycles times
    phivals_hires = np.linspace(0, cycles * yperiod, num=y_res * cycles, endpoint=False)

    xpos = np.asarray(xpos)
    zpos = np.asarray(zpos)

    field_tracer = fieldtracer.FieldTracer(magnetic_field)
    result_hires = field_tracer.follow_field_lines(xpos, zpos, phivals_hires)

    # Get phivals_hires into [0,yperiod]
    phivals_hires_mod = np.remainder(phivals_hires, yperiod)
    # There are cycles sets of field lines y_res points long each
    # and we also need to transpose for reasons
    phivals_hires_mod = phivals_hires_mod.reshape((cycles, y_res)).T
    # Same for the result, but only swap first and second indices
    result_hires_mod = result_hires.reshape((cycles, y_res, 2)).transpose(1, 0, 2)

    fig = plt.figure()
    ax = fig.add_subplot(projection="3d")
    for n in range(cycles):
        ax.plot(
            result_hires_mod[:, n, 0],
            result_hires_mod[:, n, 1],
            phivals_hires_mod[:, n],
        )

    plt.show()

    return fig, ax


def plot_streamlines(grid, magnetic_field, y_slice=0, width=None, **kwargs):
    """Plot streamlines of the magnetic field in the poloidal plane

    Parameters
    ----------
    grid : :py:obj:`zoidberg.grid.Grid`
        Grid generated by Zoidberg
    magnetic_field : :py:obj:`zoidberg.field.MagneticField`
        Zoidberg magnetic field object
    y_slice : int, optional
        y-index to plot streamlines at
    width : float, optional
        If not None, line widths are proportional to the magnitude of
        the `magnetic_field` times `width`

    Returns
    -------
    fig, ax
        The matplotlib figure and axis used

    """

    if not plotting_available:
        warnings.warning("matplotlib not available, unable to plot")
        return

    fig, ax = plt.subplots(1, 1)
    full_slice = np.s_[:, y_slice, :]

    if width is not None:
        # Get the B field magnitude in the poloidal plane
        bxz_mag = np.sqrt(magnetic_field.b_mag**2 - magnetic_field.by**2)
        linewidth = width * (bxz_mag[full_slice] / bxz_mag.max()).T
    else:
        linewidth = 1

    ax.streamplot(
        grid.xarray,
        grid.zarray,
        magnetic_field.bx[full_slice].T,
        magnetic_field.bz[full_slice].T,
        linewidth=linewidth,
        **kwargs,
    )

    ax.set_xlabel("Radius [m]", fontsize=20)
    ax.set_ylabel("Height [m]", fontsize=20)
    ax.tick_params(axis="both", labelsize=15)

    plt.show()

    return fig, ax


class AnimateVectorField(object):
    """Very basic/experimental class for animating vector fields

    Transpose U, V to have dimension to animate at index 0,
    e.g. to animate along y, pass:

    >>> AnimateVectorField(X, Y, U.transpose((1,0,2)), V.transpose((1,0,2)))

    Parameters
    ----------
    X, Y : array_like
        The X, Y coordinates
    U, V : ndarray
        Vector components in X, Y respectively

    Examples
    --------

    >>> anim = AnimateVectorField(X, Y, U, V)
    >>> anim.animate()
    """

    def __init__(self, X, Y, U, V):
        self.X = X
        self.Y = Y
        self.U = U
        self.V = V

        self.frames = U.shape[0]

        self.fig, self.ax = plt.subplots(1, 1)
        self.quiv_plot = self.ax.quiver(
            X, Y, U[0, ...].T, V[0, ...].T, pivot="mid", angles="xy"
        )

    def __update_quiver(self, num):
        self.quiv_plot.set_UVC(self.U[num, ...].T, self.V[num, ...].T)
        return (self.quiv_plot,)

    def animate(self):
        self.animation = anim.FuncAnimation(
            self.fig, self.__update_quiver, frames=self.frames, interval=100, blit=False
        )
        plt.show()
        plt.draw()


def plot_forward_map(grid, maps, yslice=0):
    """Plots the forward map from `yslice` to `yslice` + 1

    Parameters
    ----------
    grid : :py:obj:`zoidberg.grid.Grid`
        Grid generated by Zoidberg
    maps : dict
        Dictionary containing the forward FCI maps
    y_slice : int, optional
        Originating y-index to plot map from

    TODO
    ----
    - Take optional axis
    - Take optional show argument

    """

    if not plotting_available:
        warnings.warning("matplotlib not available, unable to plot")
        return

    nx, ny, nz = grid.shape

    pol, ycoord = grid.getPoloidalGrid(yslice)
    pol_next, ycoord_next = grid.getPoloidalGrid(yslice + 1)

    # Plot the points on yslice+1 as 'bx'
    # Note: ravel used here only so multiple labels are not created
    plt.plot(
        np.ravel(pol_next.R),
        np.ravel(pol_next.Z),
        "bx",
        label="Grid points on slice {0}".format(yslice + 1),
    )

    # Plot the forward map from yslice to yslice+1 as red 'o'
    forward_R = maps["forward_R"][:, yslice, :]
    forward_Z = maps["forward_Z"][:, yslice, :]
    plt.plot(
        np.ravel(forward_R),
        np.ravel(forward_Z),
        "ro",
        label="Forward map from slice {0}".format(yslice),
    )

    # Mark the points which hit the inner boundary
    # These are marked with a negative x index
    in_boundary = maps["forward_xt_prime"][:, yslice, :] < 0.0
    plt.plot(
        np.ravel(forward_R[in_boundary]),
        np.ravel(forward_Z[in_boundary]),
        "ko",
        label="Inner boundary points",
    )

    # Outer boundary marked with x index nx
    out_boundary = maps["forward_xt_prime"][:, yslice, :] > nx - 0.5
    plt.plot(
        np.ravel(forward_R[out_boundary]),
        np.ravel(forward_Z[out_boundary]),
        "bo",
        label="Outer boundary points",
    )

    plt.legend()

    plt.show()


def plot_backward_map(grid, maps, yslice=0):
    """Plots the backward map from yslice to yslice-1

    Parameters
    ----------
    grid : 'zoidberg.grid.Grid`
        Grid generated by Zoidberg
    maps : dict
        Dictionary containing the backward FCI maps
    y_slice : int, optional
        Originating y-index to plot map from

    TODO
    ----
    - Take optional axis
    - Take optional show argument

    """

    if not plotting_available:
        warnings.warning("matplotlib not available, unable to plot")
        return

    nx, ny, nz = grid.shape

    pol, ycoord = grid.getPoloidalGrid(yslice)
    pol_last, ycoord_last = grid.getPoloidalGrid(yslice - 1)

    # Plot the points on yslice-1 as 'bx'
    # Note: ravel used here only so multiple labels are not created
    plt.plot(
        np.ravel(pol_last.R),
        np.ravel(pol_last.Z),
        "bx",
        label="Grid points on slice {0}".format(yslice - 1),
    )

    # Plot the backward map from yslice to yslice-1 as red 'o'
    backward_R = maps["backward_R"][:, yslice, :]
    backward_Z = maps["backward_Z"][:, yslice, :]
    plt.plot(
        np.ravel(backward_R),
        np.ravel(backward_Z),
        "ro",
        label="Backward map from slice {0}".format(yslice),
    )

    # Mark the points which hit the inner boundary
    # These are marked with a negative x index
    in_boundary = maps["backward_xt_prime"][:, yslice, :] < 0.0
    plt.plot(
        np.ravel(backward_R[in_boundary]),
        np.ravel(backward_Z[in_boundary]),
        "ko",
        label="Inner boundary points",
    )

    # Outer boundary marked with x index nx
    out_boundary = maps["backward_xt_prime"][:, yslice, :] > nx - 0.5
    plt.plot(
        np.ravel(backward_R[out_boundary]),
        np.ravel(backward_Z[out_boundary]),
        "bo",
        label="Outer boundary points",
    )

    plt.legend()

    plt.show()



#!/usr/bin/env python
# coding: utf-8

import time

time_start = time.time()  # noqa

import sys

import numpy as np
from boututils.datafile import DataFile as DF

verbose = 1


def toCent(RZ):
    RZc = RZ + np.roll(RZ, 1, axis=-1)
    RZc = RZc[:, 1:] + RZc[:, :-1]
    return RZc / 4


def l2(x):
    return np.sqrt(np.sum(x**2, axis=0))


org_print = print


def my_print(*args):
    args = f"{time.time() - time_start:10.3f} s: ", *args
    org_print(*args)


def log(*args):
    if verbose > 1:
        my_print(*args)


def print(*args):
    if verbose:
        my_print(*args)


def load(fn):
    with DF(fn) as f:
        RZ = [f[k] for k in "RZ"]
        log("opened")
        RZ = np.array(RZ)
        _, a, b, c = RZ.shape

        coefsX = np.zeros((a - 1, b, c, 2))
        coefsZ = np.zeros((a - 2, b, c, 2))
        for d, x1 in zip((coefsX, coefsZ), "XZ"):
            for i, x2 in enumerate("XZ"):
                key = f"dagp_fv_{x1}{x2}"
                fixup2(d[..., i], f[key])
        key = "dagp_fv_volume"
        volume = f[key]
    log("done reading")
    return RZ, volume, coefsX, coefsZ


def doit(pols, plot=False):
    RZs = []

    ### Calculate Volume of the cell
    #
    ### Go in a line around the cell

    A = np.empty((pols[0].nx, len(pols), pols[0].nz))
    Ar = np.empty_like(A)

    for gi, g in enumerate(pols):
        n = 50
        x0 = np.arange(g.nx)[1:-1, None, None]
        y0 = np.arange(g.nz)[None, :, None]
        x1 = x0[:, :, 0]
        y1 = y0[:, :, 0]
        x = []
        y = []
        zero = np.zeros((g.nx - 2, g.nz, n))
        x += [np.linspace(x1 - 0.5, x1 + 0.5, n).transpose(1, 2, 0)]
        y += [y0 + 0.5]
        x += [x0 + 0.5]
        y += [np.linspace(y1 + 0.5, y1 - 0.5, n).transpose(1, 2, 0)]
        x += [np.linspace(x1 + 0.5, x1 - 0.5, n).transpose(1, 2, 0)]
        y += [y0 - 0.5]
        x += [x0 - 0.5]
        y += [np.linspace(y1 - 0.5, y1 + 0.5, n).transpose(1, 2, 0)]
        x = np.concatenate([k + zero for k in x], axis=-1)
        y = np.concatenate([k + zero for k in y], axis=-1)
        RZ = np.array(g.getCoordinate(x, y))

        dy = RZ[1, ..., 1:] - RZ[1, ..., :-1]
        xx = (RZ[0, ..., :-1] + RZ[0, ..., 1:]) / 2

        A[1:-1, gi] = -np.sum(xx * dy, axis=-1)
        Ar[1:-1, gi] = -np.sum(0.5 * xx * xx * dy, axis=-1)

        RZs += [RZ]
    RZs = np.array(RZs)
    print(RZs.shape)
    # 3.160 s:  (4, 2, 6, 32, 200)
    RZs = RZs.transpose(1, 2, 0, 3, 4)
    print(RZs.shape)

    volume = Ar
    # f =
    #    V = [f[k] for k in ("dx", "dy", "dz", "J")]

    RZ = np.array(
        [
            g.getCoordinate(
                *np.meshgrid(np.arange(g.nx), np.arange(g.nz), indexing="ij")
            )
            for g in pols
        ]
    ).transpose(1, 2, 0, 3)

    RZ = np.array([(g.R, g.Z) for g in pols]).transpose(1, 2, 0, 3)

    # volume = dx * dy * dz * J
    # volume = V[0] * V[1] * V[2] * V[3]
    # A.shape, Ar.shape, RZ.shape

    # AreaXplus
    # Z
    # Λ
    # |   a        b          c
    # |
    # |       ------------
    # |       |          X
    # |       |          X
    # |   h   |    o     X    d
    # |       |          X
    # |       |          X
    # |       ------------
    # |
    # |   g        f          e
    # |
    # |------------------------> x
    #
    # We are calculating the area of the surface denoted by X
    # Note that we need to split it in two parts.

    log("calculating pos ...")
    # pos = RZs[..., :n]
    tmp = np.meshgrid(
        np.arange(g.nx - 1) + 0.5,
        np.arange(g.nz) - 0.5,
        np.linspace(0, 1, n),
        indexing="ij",
    )
    tmp[1] += tmp[2]
    print(tmp[0])
    print(tmp[1])
    pos = np.array([g.getCoordinate(*tmp[:2]) for g in pols]).transpose(1, 2, 0, 3, 4)
    log("done")
    # direction of edge, length ~ 1/nx
    dR = pos[..., :-1] - pos[..., 1:]
    dX = np.sqrt(np.sum(dR**2, axis=0))
    area = dX * (pos[0, ..., 1:] + pos[0, ..., :-1]) * 0.5
    assert area.shape[0] > 2
    # Vector in normal direction
    assert dR.shape[0] == 2
    dRr = np.array((-dR[1], dR[0]))
    dRr /= np.sqrt(np.sum(dRr**2, axis=0))
    dRr *= area
    dRr = np.sum(dRr, axis=-1)
    print(np.nanmean(dRr))

    # vector in derivative direction
    dxR = RZ[:, 1:] - RZ[:, :-1]
    print("Maximum radial grid distance:", np.max(l2(dxR)))
    dzR = np.roll(RZ, -1, axis=-1) - np.roll(RZ, 1, axis=-1)
    dzR = 0.5 * (dzR[:, 1:] + dzR[:, :-1])

    # dxR /= (np.sum(dxR**2, axis=0))
    # dzR /= (np.sum(dzR**2, axis=0))
    log("starting solve")
    dxzR = np.array((dxR, dzR)).transpose(2, 3, 4, 1, 0)
    coefsX = np.linalg.solve(dxzR, dRr.transpose(1, 2, 3, 0))
    log("done")

    # AreaZplus
    # Z
    # Λ
    # |   a        b          c
    # |
    # |       -XXXXXXXXXX-
    # |       |          |
    # |       |          |
    # |   h   |    o     |    d
    # |       |          |
    # |       |          |
    # |       ------------
    # |
    # |   g        f          e
    # |
    # |------------------------> x

    log("concatenate")
    tmp = np.meshgrid(
        np.arange(g.nx - 2) + 0.5,
        np.arange(g.nz) + 0.5,
        np.linspace(0, 1, n),
        indexing="ij",
    )
    tmp[0] += tmp[2]
    pos = np.array([g.getCoordinate(*tmp[:2]) for g in pols]).transpose(1, 2, 0, 3, 4)

    dR = pos[..., :-1] - pos[..., 1:]
    dX = np.sqrt(np.sum(dR**2, axis=0))
    area = dX * (pos[0, ..., 1:] + pos[0, ..., :-1]) * 0.5
    log("get normal vector")
    # Vector in normal direction
    dRr = np.array((dR[1], -dR[0]))
    dRr /= np.sqrt(np.sum(dRr**2, axis=0))
    dRr *= area
    dRr = np.sum(dRr, axis=-1)
    # vector in derivative direction
    dxR = RZ[:, 2:] - RZ[:, :-2]
    dxR = 0.5 * (np.roll(dxR, -1, axis=-1) + dxR)
    dzR = (np.roll(RZ, -1, axis=-1) - RZ)[:, 1:-1]
    dxzR = np.array((dxR, dzR)).transpose(2, 3, 4, 1, 0)
    log("solving again")
    coefsZ = np.linalg.solve(dxzR, dRr.transpose(1, 2, 3, 0))
    log("done")

    test(RZ, volume, coefsX, coefsZ, plot=plot)

    return write(RZ, volume, coefsX, coefsZ)


def test(RZ, volume, coefsX, coefsZ, plot=False):
    inp = np.sin(RZ[1])
    ana = -np.sin(RZ[1])
    if 0:
        inp = RZ[1] ** 2
        ana = RZ[1] ** 0
    if 1:
        inp = np.sin(RZ[0])
        ana = -np.sin(RZ[0]) + np.cos(RZ[0]) / RZ[0]
    if 0:
        inp = RZ[0] ** 3
        ana = 6 * np.sin(RZ[0])

    def xp(ijk):
        return (ijk[0] + 1, *ijk[1:])

    def xm(ijk):
        return (ijk[0] - 1, *ijk[1:])

    def zp(ijk):
        return _zp(*ijk)

    def _zp(i, j, k):
        return (i, j, (k + 1) % zmax)

    def zm(i, j, k):
        return (i, j, (k - 1) % zmax)

    log("testing")
    result = np.zeros(inp.shape)
    results = [np.zeros(inp.shape) for _ in range(4)]
    # dx =
    dz2 = np.roll(inp, -1, axis=-1) - np.roll(inp, 1, axis=-1)
    i, j, k = inp.shape
    zmax = k
    dx = inp[1:] - inp[:-1]
    dz = 0.5 * (dz2[1:] + dz2[:-1])
    this = -(coefsX[..., 0] * dx + coefsX[..., 1] * dz)
    result[:-1] -= this
    result[1:] += this
    for r, t in zip(results, (-coefsX[..., 0] * dx, -coefsX[..., 1] * dz)):
        r[:-1] -= t
        r[1:] += t
    print("expect 0:", np.max(np.abs((result - results[0] - results[1]))))
    if 1:
        dx2 = inp[2:] - inp[:-2]
        dx2 = 0.5 * (np.roll(dx2, -1, axis=-1) + dx2)
        dz2 = (np.roll(inp, -1, axis=-1) - inp)[1:-1]
        t1 = coefsZ[..., 0] * dx2
        t2 = coefsZ[..., 1] * dz2
        this = -(t1 + t2)
        result[1:-1] -= this
        result[1:-1] += np.roll(this, 1, -1)
        for r, t in zip(results[2:], (-t1, -t2)):
            r[1:-1] -= t
            # np.roll(r, -1, -1)[1:-1] += t
            r[1:-1] += np.roll(t, 1, -1)

    result[0] = 0
    result[-1] = 0
    for r in results:
        r[0] = 0
        r[-1] = 0

    if plot:
        import matplotlib.pyplot as plt

        f, axs = plt.subplots(1, 3, sharex=True, sharey=True)
        res = result / volume
        for d, ax, t in zip(
            (res, ana, res - ana),
            axs,
            ("computed", "analytical", "error"),
        ):
            slc = slice(1, -1), 1
            per = np.percentile(d[slc], [0, 2, 98, 100])
            print(per)
            p = ax.pcolormesh(*[k[slc] for k in RZ], d[slc], vmin=per[1], vmax=per[2])
            ax.set_title(t)
            plt.colorbar(p, ax=ax)
        plt.show()
    print("error:", np.mean(l2(result[1:-1] / volume[1:-1] - ana[1:-1])))


def fixup(RZ, d):
    c1 = np.zeros(RZ[0].shape)
    if c1.shape[0] == d.shape[0] + 1:
        c1[:-1] = d
    else:
        c1[1:-1] = d
    # c1 = np.roll(c1, -1, -1)
    return c1


def fixup2(d, val):
    if val.shape[0] == d.shape[0] + 1:
        d[:] = val[:-1]
    else:
        d[:] = val[1:-1]


def write(RZ, volume, coefsX, coefsZ):
    ret = {}
    log("writing")
    for d, x1 in zip((coefsX, coefsZ), "XZ"):
        for i, x2 in enumerate("XZ"):
            key = f"dagp_fv_{x1}{x2}"
            ret[key] = fixup(RZ, d[..., i])
        key = "dagp_fv_volume"
        ret[key] = volume
    log("done")
    print(ret.keys())
    return ret


if __name__ == "__main__":
    for flag, step in (("-v", +1), ("-q", -1)):
        while flag in sys.argv:
            verbose += step
            sys.argv.remove(flag)
    plot = "-p" in sys.argv
    if plot:
        sys.argv.remove("-p")

    for fn in sys.argv[1:]:
        print(fn)
        test(*load(fn), plot=plot)



import uuid
from collections import namedtuple
from itertools import chain

import numpy as np
from boututils import datafile as bdata

from zoidberg import __version__

from . import fieldtracer
from .grid import Grid
from .poloidal_grid import StructuredPoloidalGrid

try:
    from tqdm.auto import tqdm
except ModuleNotFoundError:
    tqdm = None
    from .progress import update_progress

# PyEVTK might be called pyevtk or evtk, depending on where it was
# installed from
have_evtk = True
try:
    from pyevtk.hl import gridToVTK
except ImportError:
    try:
        from evtk.hl import gridToVTK
    except ImportError:
        have_evtk = False


def parallel_slice_field_name(field, offset):
    """Form a unique, backwards-compatible name for field at a given offset

    Parameters
    ----------
    field : str
        Name of the field to convert
    offset : int
        Parallel slice offset

    """
    prefix = "forward" if offset > 0 else "backward"
    suffix = "_{}".format(abs(offset)) if abs(offset) > 1 else ""
    return "{}_{}{}".format(prefix, field, suffix)


def make_maps(grid, magnetic_field, nslice=1, quiet=False, field_tracer=None, **kwargs):
    """Make the forward and backward FCI maps

    Parameters
    ----------
    grid : :py:obj:`zoidberg.grid.Grid`
        Grid generated by Zoidberg
    magnetic_field : :py:obj:`zoidberg.field.MagneticField`
        Zoidberg magnetic field object
    nslice : int
        Number of parallel slices in each direction
    quiet : bool
        Don't display progress bar
    kwargs
        Optional arguments for field line tracing, etc.

    Returns
    -------
    dict
        Dictionary containing the forward/backward field line maps

    """

    # Get number of points
    # Note: Assumes that all poloidal grids have the same number of x and z(y) points
    ny = grid.numberOfPoloidalGrids()
    pol, _ = grid.getPoloidalGrid(0)
    nx = pol.nx
    nz = pol.nz

    # Get number of guard cells (default 2)
    mxg = kwargs.get("MXG", 2)

    shape = (nx, ny, nz)

    # Coordinates of each grid point
    R = np.zeros(shape)
    Z = np.zeros(shape)

    for j in range(ny):
        pol, _ = grid.getPoloidalGrid(j)
        R[:, j, :] = pol.R
        Z[:, j, :] = pol.Z

    field_tracer = field_tracer or fieldtracer.FieldTracer(magnetic_field)

    rtol = kwargs.get("rtol", None)

    # The field line maps and coordinates, etc.
    maps = {
        "R": R,
        "Z": Z,
        "MXG": mxg,
        "MYG": nslice,
    }

    # A helper data structure that groups the various field line maps along with the offset
    ParallelSlice = namedtuple(
        "ParallelSlice", ["offset", "R", "Z", "xt_prime", "zt_prime"]
    )
    # A list of the above data structures for each offset we want
    parallel_slices = []

    # Loop over offsets {1, ... nslice, -1, ... -nslice}
    for offset in chain(range(1, nslice + 1), range(-1, -(nslice + 1), -1)):
        # Unique names of the field line maps for this offset
        field_names = [
            parallel_slice_field_name(field, offset)
            for field in ["R", "Z", "xt_prime", "zt_prime"]
        ]

        # Initialise the field arrays -- puts them straight into the result dict
        for field in field_names:
            maps[field] = np.zeros(shape)

        # Get the field arrays we just made and wrap them up in our helper tuple
        fields = map(lambda x: maps[x], field_names)
        parallel_slices.append(ParallelSlice(offset, *fields))

    # Total size of the progress bar
    total_work = len(parallel_slices) * ny

    # TODO: if axisymmetric, don't loop, do one slice and copy
    # TODO: restart tracing for adjacent offsets
    if (not quiet) and (ny > 1):
        if tqdm:
            prog = tqdm(total=total_work, desc="Tracing")
        else:
            update_progress(0, **kwargs)
    for slice_index, parallel_slice in enumerate(parallel_slices):
        for j in range(ny):
            # Get this poloidal grid
            pol, ycoord = grid.getPoloidalGrid(j)

            # Get the next poloidal grid
            pol_slice, y_slice = grid.getPoloidalGrid(j + parallel_slice.offset)

            # We only want the end point, as [0,...] is the initial position
            coord = field_tracer.follow_field_lines(
                pol.R, pol.Z, [ycoord, y_slice], rtol=rtol
            )[1, ...]

            # Store the coordinates in real space
            parallel_slice.R[:, j, :] = coord[:, :, 0]
            parallel_slice.Z[:, j, :] = coord[:, :, 1]

            # Get the indices into the slice poloidal grid
            if pol_slice is None:
                # No slice grid, so hit a boundary
                xind = -1
                zind = -1
            else:
                # Find the indices for these new locations on the slice poloidal grid
                xcoord = coord[:, :, 0]
                zcoord = coord[:, :, 1]
                xind, zind = pol_slice.findIndex(xcoord, zcoord)

                # Check boundary defined by the field
                outside = magnetic_field.boundary.outside(xcoord, y_slice, zcoord)
                xind[outside] = -1
                zind[outside] = -1

            parallel_slice.xt_prime[:, j, :] = xind
            parallel_slice.zt_prime[:, j, :] = zind

            if (not quiet) and (ny > 1):
                if tqdm:
                    prog.update()
                else:
                    update_progress((slice_index * ny + j + 1) / total_work, **kwargs)

    return maps


def update_metric_names(metric):
    # Translate between output variable names and metric names
    # Map from new to old names. Anything not in this dict
    # is unchanged
    name_changes = {
        "g_yy": "g_22",
        "gyy": "g22",
        "gxx": "g11",
        "gxz": "g13",
        "gzz": "g33",
        "g_xx": "g_11",
        "g_xz": "g_13",
        "g_zz": "g_33",
    }
    return {name_changes.get(key, key): value for key, value in metric.items()}


def get_metric(grid, magnetic_field):
    nx, ny, nz = grid.shape
    # Get metric tensor
    metric = grid.metric()

    # Check if the magnetic field is in cylindrical coordinates
    # If so, we need to change the gyy and g_yy metrics
    pol_grid, ypos = grid.getPoloidalGrid(0)
    Rmaj = magnetic_field.Rfunc(pol_grid.R, pol_grid.Z, ypos)
    if Rmaj is not None:
        # In cylindrical coordinates
        Rmaj = np.zeros(grid.shape)
        for yindex in range(grid.numberOfPoloidalGrids()):
            pol_grid, ypos = grid.getPoloidalGrid(yindex)
            Rmaj[:, yindex, :] = magnetic_field.Rfunc(pol_grid.R, pol_grid.Z, ypos)
        metric["gyy"] = 1.0 / Rmaj**2
        metric["g_yy"] = Rmaj**2

    # Get magnetic field and pressure
    Bmag = np.zeros(grid.shape)
    pressure = np.zeros(grid.shape)
    print("starting Bfield stuff")
    for yindex in range(grid.numberOfPoloidalGrids()):
        pol_grid, ypos = grid.getPoloidalGrid(yindex)
        Bmag[:, yindex, :] = magnetic_field.Bmag(pol_grid.R, pol_grid.Z, ypos)
        pressure[:, yindex, :] = magnetic_field.pressure(pol_grid.R, pol_grid.Z, ypos)
        By = magnetic_field.Byfunc(pol_grid.R, pol_grid.Z, ypos)
        metric["g_yy"][:, yindex, :] *= (Bmag[:, yindex, :] / By) ** 2
        metric["gyy"][:, yindex, :] *= (By / Bmag[:, yindex, :]) ** 2
    print("done Bfield stuff")

    return metric, Bmag, pressure


def write_maps(
    grid,
    magnetic_field,
    maps,
    gridfile="fci.grid.nc",
    new_names=False,
    metric2d=True,
    format="NETCDF4",
    quiet=False,
):
    """Write FCI maps to BOUT++ grid file

    Parameters
    ----------
    grid : :py:obj:`zoidberg.grid.Grid`
        Grid generated by Zoidberg
    magnetic_field : :py:obj:`zoidberg.field.MagneticField`
        Zoidberg magnetic field object
    maps : dict
        Dictionary of FCI maps
    gridfile : str, optional
        Output filename
    new_names : bool, optional
        Write "g_yy" rather than "g_22"
    metric2d : bool, optional
        Output only 2D metrics
    format : str, optional
        Specifies file format to use, passed to boutdata.DataFile
    quiet : bool, optional
        Don't warn about 2D metrics

    Returns
    -------

    Writes the following variables to the grid file


    """
    metric, Bmag, pressure = get_metric(grid, magnetic_field)

    nx, ny, nz = grid.shape

    # Add Rxy, Bxy
    metric["Rxy"] = maps["R"]
    metric["Bxy"] = Bmag

    nslice = maps["MYG"]
    # Loop over offsets {1, ... nslice, -1, ... -nslice}
    for offset in chain(range(1, nslice + 1), range(-1, -(nslice + 1), -1)):
        par_pgrids, ypar = np.array(
            [grid.getPoloidalGrid(i + offset) for i in range(ny)], dtype=object
        ).T
        par_pgrids = [StructuredPoloidalGrid(p.R, p.Z) for p in par_pgrids]
        par_grid = Grid(
            par_pgrids,
            ypar,
            Ly=grid.Ly,
            yperiodic=grid.yperiodic,
        )

        par_metric, _, _ = get_metric(par_grid, magnetic_field)
        if not new_names:
            par_metric = update_metric_names(par_metric)

        for k, v in par_metric.items():
            name = parallel_slice_field_name(k, offset)
            assert name not in metric
            metric[name] = v

    # Get attributes from magnetic field (e.g. psi)
    attributes = {}
    for name in magnetic_field.attributes:
        attribute = np.zeros(grid.shape)
        for yindex in range(grid.numberOfPoloidalGrids()):
            pol_grid, ypos = grid.getPoloidalGrid(yindex)
            attribute[:, yindex, :] = magnetic_field.attributes[name](
                pol_grid.R, pol_grid.Z, ypos
            )
            attributes[name] = attribute

    # Metric is now 3D
    if metric2d:
        # Remove the Z dimension from metric components
        if not quiet:
            print("WARNING: Outputting 2D metrics, discarding metric information.")
        for key in metric:
            try:
                metric[key] = metric[key][:, :, 0]
            except TypeError:
                pass
        # Make dz a constant
        metric["dz"] = metric["dz"][0, 0]

    if not new_names:
        metric = update_metric_names(metric)

    with bdata.DataFile(gridfile, write=True, create=True, format=format) as f:
        f.write_file_attribute("title", "BOUT++ grid file")
        f.write_file_attribute("software_name", "zoidberg")
        f.write_file_attribute("software_version", __version__)
        grid_id = str(uuid.uuid1())
        f.write_file_attribute("id", grid_id)  # conventional name
        f.write_file_attribute("grid_id", grid_id)  # BOUT++ specific name

        ixseps = nx + 1
        f.write("nx", nx)
        f.write("ny", ny)
        f.write("nz", nz)

        f.write("dx", metric["dx"])
        f.write("dy", metric["dy"])
        f.write("dz", metric["dz"])

        f.write("ixseps1", ixseps)
        f.write("ixseps2", ixseps)

        # Metric tensor
        for key in metric:
            f.write(key, metric[key])

        # Magnetic field
        f.write("B", Bmag)

        # Pressure
        f.write("pressure", pressure)

        # Attributes
        for name in attributes:
            f.write(name, attributes[name])

        # Maps - write everything to file
        for key in maps:
            f.write(key, maps[key])


def write_Bfield_to_vtk(
    grid, magnetic_field, scale=5, vtkfile="fci_zoidberg", psi=True
):
    """Write the magnetic field to a VTK file

    Parameters
    ----------
    grid : :py:obj:`zoidberg.grid.Grid`
        Grid generated by Zoidberg
    magnetic_field : :py:obj:`zoidberg.field.MagneticField`
        Zoidberg magnetic field object
    scale : int, optional
        Factor to scale x, z dimensions by [5]
    vtkfile : str, optional
        Output filename without extension ["fci_zoidberg"]
    psi : bool, optional
        Write psi?

    Returns
    -------
    path           - Full path to vtkfile
    """

    point_data = {
        "B": (magnetic_field.bx * scale, magnetic_field.by, magnetic_field.bz * scale)
    }

    if psi:
        psi = make_surfaces(grid, magnetic_field)
        point_data["psi"] = psi

    path = gridToVTK(
        vtkfile,
        grid.xarray * scale,
        grid.yarray,
        grid.zarray * scale,
        pointData=point_data,
    )

    return path


def fci_to_vtk(infile, outfile, scale=5):
    if not have_evtk:
        return

    with bdata.DataFile(infile, write=False, create=False) as f:
        dx = f.read("dx")
        dy = f.read("dy")

        bx = f.read("bx")
        by = np.ones(bx.shape)
        bz = f.read("bz")
        if bx is None:
            xt_prime = f.read("forward_xt_prime")
            zt_prime = f.read("forward_zt_prime")
            array_indices = np.indices(xt_prime.shape)
            bx = xt_prime - array_indices[0, ...]
            by = by * dy
            bz = zt_prime - array_indices[2, ...]

        nx, ny, nz = bx.shape
        dz = nx * dx / nz

    x = np.linspace(0, nx * dx, nx)
    y = np.linspace(0, ny * dy, ny, endpoint=False)
    z = np.linspace(0, nz * dz, nz, endpoint=False)

    gridToVTK(
        outfile, x * scale, y, z * scale, pointData={"B": (bx * scale, by, bz * scale)}
    )


def make_surfaces(grid, magnetic_field, nsurfaces=10, revs=100):
    """Essentially interpolate a poincare plot onto the grid mesh

    Parameters
    ----------
    grid : :py:obj:`zoidberg.grid.Grid`
        Grid generated by Zoidberg
    magnetic_field : :py:obj:`zoidberg.field.MagneticField`
        Zoidberg magnetic field object
    nsurfaces : int, optional
        Number of surfaces to interpolate to [10]
    revs : int, optional
        Number of points on each surface [100]

    Returns
    -------
    surfaces
        Array of psuedo-psi on the grid mesh

    """

    from scipy.interpolate import griddata

    # initial x, z points in surface
    xpos = grid.xcentre + np.linspace(0, 0.5 * np.max(grid.xarray), nsurfaces)
    zpos = grid.zcentre

    phi_values = grid.yarray[:]
    # Extend the domain from [0,grid.Ly] to [0,revs*grid.Ly]
    for n in np.arange(1, revs):
        phi_values = np.append(phi_values, n * grid.Ly + phi_values[: grid.ny])

    # Get field line tracer and trace out surfaces
    tracer = fieldtracer.FieldTracer(magnetic_field)
    points = tracer.follow_field_lines(xpos, zpos, phi_values)

    # Reshape to be easier to work with
    points = points.reshape((revs, grid.ny, nsurfaces, 2))

    # Arbitarily number the surfaces from 0 to 1
    psi_points = np.zeros((revs, grid.ny, nsurfaces))
    for surf in range(nsurfaces):
        psi_points[:, :, surf] = float(surf) / float(nsurfaces - 1)

    x_2d, z_2d = np.meshgrid(grid.xarray, grid.zarray, indexing="ij")

    psi = np.zeros_like(grid.x_3d)
    for y_slice in range(grid.ny):
        points_2d = np.column_stack(
            (points[:, y_slice, :, 0].flatten(), points[:, y_slice, :, 1].flatten())
        )
        psi[:, y_slice, :] = griddata(
            points_2d,
            psi_points[:, y_slice, :].flatten(),
            (x_2d, z_2d),
            method="linear",
            fill_value=1,
        )

    return psi


def upscale(field, maps, upscale_factor=4, quiet=True):
    """Increase the resolution in y of field along the FCI maps.

    First, interpolate onto the (forward) field line end points, as in
    normal FCI technique. Then interpolate between start and end
    points. We also need to interpolate the xt_primes and
    zt_primes. This gives a cloud of points along the field lines,
    which we can finally interpolate back onto a regular grid.

    Parameters
    ----------
    field : array_like
        3D field to be upscaled
    maps : dict
        Zoidberg field line maps
    upscale_factor : int, optional
        Factor to increase resolution by [4]
    quiet : bool, optional
        Don't show progress bar [True]

    Returns
    -------
    Field with y-resolution increased *upscale_factor* times. Shape is
    (nx, upscale_factor*ny, nz).

    """

    from scipy.interpolate import griddata
    from scipy.ndimage.interpolation import map_coordinates

    xt_prime = maps["forward_xt_prime"]
    zt_prime = maps["forward_zt_prime"]

    # The field should be the same shape as the grid
    if field.shape != xt_prime.shape:
        try:
            field = field.reshape(xt_prime.T.shape).T
        except ValueError:
            raise ValueError(
                "Field, {}, must be same shape as grid, {}".format(
                    field.shape, xt_prime.shape
                )
            )

    # Get the shape of the grid
    nx, ny, nz = xt_prime.shape
    index_coords = np.mgrid[0:nx, 0:ny, 0:nz]

    # We use the forward maps, so get the y-index of the *next* y-slice
    yup_3d = index_coords[1, ...] + 1
    yup_3d[:, -1, :] = 0

    # Index space coordinates of the field line end points
    end_points = np.array([xt_prime, yup_3d, zt_prime])

    # Interpolation of the field at the end points
    field_prime = map_coordinates(field, end_points)

    # This is a 4D array where the first dimension is the start/end of
    # the field line
    field_aligned = np.array([field, field_prime])

    # x, z coords at start/end of field line
    x_start_end = np.array([index_coords[0, ...], xt_prime])
    z_start_end = np.array([index_coords[2, ...], zt_prime])

    # Parametric points along the field line
    midpoints = np.linspace(0, 1, upscale_factor, endpoint=False)
    # Need to make this 4D as well
    new_points = np.tile(midpoints[:, np.newaxis, np.newaxis, np.newaxis], [nx, ny, nz])

    # Index space coordinates of our upscaled field
    index_4d = np.mgrid[0:upscale_factor, 0:nx, 0:ny, 0:nz]
    hires_points = np.array(
        [new_points, index_4d[1, ...], index_4d[2, ...], index_4d[3, ...]]
    )

    # Upscale the field
    hires_field = map_coordinates(field_aligned, hires_points)

    # Linearly interpolate the x, z coordinates of the field lines
    hires_x = map_coordinates(x_start_end, hires_points)
    hires_z = map_coordinates(z_start_end, hires_points)

    def twizzle(array):
        """Transpose and reshape the output of map_coordinates to
        be 3D
        """
        return array.transpose((1, 2, 0, 3)).reshape((nx, upscale_factor * ny, nz))

    # Rearrange arrays to be 3D
    hires_field = twizzle(hires_field)
    hires_x = twizzle(hires_x)
    hires_z = twizzle(hires_z)

    # Interpolate from field line sections onto grid
    hires_grid_field = np.zeros((nx, upscale_factor * ny, nz))
    hires_index_coords = np.mgrid[0:nx, 0 : ny : 1.0 / upscale_factor, 0:nz]
    grid_points = (hires_index_coords[0, :, 0, :], hires_index_coords[2, :, 0, :])

    def y_first(array):
        """Put the middle index first"""
        return array.transpose((0, 2, 1))

    # The hires data is unstructed only in (x,z), interpolate onto
    # (x,z) grid for each y-slice individually
    for k, (x_points, z_points, f_slice) in enumerate(
        zip(y_first(hires_x).T, y_first(hires_z).T, y_first(hires_field).T)
    ):
        points = np.column_stack((x_points.flat, z_points.flat))
        hires_grid_field[:, k, :] = griddata(
            points, f_slice.flat, grid_points, method="linear", fill_value=0.0
        )
        if not quiet:
            update_progress(float(k) / float(ny - 1))

    return hires_grid_field



import numpy as np

from . import diff


def l2(f):
    return np.sqrt(np.mean(f**2))


def mms(testfunc, inp, ana, xmax=2 * np.pi, slices=None):
    if slices is None:
        slices = [slice(None)]
    l2s = []
    for nx in [64, 128]:
        x = np.linspace(0, xmax, nx, endpoint=False)
        err = testfunc(inp(x)) * nx / xmax - ana(x)
        l2s.append([l2(err[s]) for s in slices])
        if xmax == 2:
            import matplotlib.pyplot as plt

            plt.plot(x, testfunc(inp(x)) * nx / xmax, label="result")
            plt.plot(x, ana(x), label="ana")
            plt.legend()
            plt.show()
    return l2s


def check_l2s(l2s, exps):
    if isinstance(exps, int):
        exps = [exps]
    l2s = np.array(l2s).T
    print(l2s.shape)
    success = True
    for (a, b), exp in zip(l2s, exps):
        ord = np.log(a / b) / np.log(2)
        s = np.isclose(ord, exp, atol=0.25)
        if not s:
            success = False
        s = "👍" if s else "❌"
        print(f"{ord:.3f} vs {exp} {s}")
    return success


def test_diff_c2_per():
    print("C2 periodic")
    l2 = mms(lambda x: diff.c2(x, 0, True), np.sin, np.cos)
    assert check_l2s(l2, 2)


def test_diff_c2_np():
    print("C2 periodic")
    l2 = mms(
        lambda x: diff.c2(x, 0, False),
        np.sin,
        np.cos,
        1,
        slices=[slice(0, 1), slice(1, -1), slice(-1)],
    )
    assert check_l2s(l2, [2, 2, 2])


def test_diff_c4_per():
    print("C4 periodic")
    l2 = mms(lambda x: diff.c4(x, 0, True), np.sin, np.cos)
    assert check_l2s(l2, 4)


def test_diff_c4_np():
    print("C4 non periodic")
    l2 = mms(
        lambda x: diff.c4(x, 0, False),
        np.sin,
        np.cos,
        1,
        slices=[slice(0, 2), slice(2, -2), slice(-2)],
    )
    assert check_l2s(l2, [4, 4, 4])

    l2 = mms(
        lambda x: diff.c4(x, 0, False),
        np.cos,
        lambda x: -np.sin(x),
        1,
    )
    assert check_l2s(l2, 4)


if __name__ == "__main__":
    test_diff_c4_np()
    test_diff_c4_per()



import numpy as np

from . import poloidal_grid, rzline


def test_out_of_domain():
    # Create inner and outer boundaries
    inner = rzline.circle(R0=0.0, r=1.0, n=20)
    outer = rzline.circle(R0=0.0, r=2.0, n=20)

    grid = poloidal_grid.grid_elliptic(inner, outer, 10, 10)

    # Test a point in domain

    x, y = grid.findIndex(1.0, 1.0)
    r, z = grid.getCoordinate(x, y)
    assert np.allclose(r, 1.0) and np.allclose(z, 1.0)  # Should get original R,Z

    # Test out of domain

    x, y = grid.findIndex(2.5, 0)
    assert np.allclose(x, 10)  # Should mark out of domain at outer x

    x, y = grid.findIndex(0.5, 0.2)
    assert np.allclose(x, -1)  # Should mark out of domain at inner x

    # Test a mix of values inside and outside domain

    x, y = grid.findIndex([2.5, 1.0], [0.0, 0.5])

    assert np.allclose(x[0], 10)  # First one out of domain

    r, z = grid.getCoordinate(x[1], y[1])
    assert np.allclose(r, 1.0) and np.allclose(z, 0.5)  # Second point in domain



import numpy as np

# Import classes representing poloidal grids
from .poloidal_grid import RectangularPoloidalGrid


class Grid(object):
    """Represents a 3D grid, consisting of a collection of poloidal grids

    Attributes
    ----------
    shape : (int, int, int)
        Tuple of grid sizes (nx, ny, nz)

    Parameters
    ----------
    poloidal_grids : :py:obj:`list` of :py:obj:`~zoidberg.poloidal_grid.PoloidalGrid`
        The collection of poloidal grids to group together
    ycoords : array_like
        The y-coordinate corresponding to each element of `poloidal_grids`

    Examples
    --------

    >>> poloidal_grids = [RectangularPoloidalGrid(5, 5, 1, 1)]
    >>> ycoords = [0.0]
    >>> grid = Grid(poloidal_grids, ycoords)

    To iterate over the poloidal grids, and get the grids to either side:

    >>> for i in range(grid.numberOfPoloidalGrids()):
    ...     pol, y = grid.getPoloidalGrid(i)
    ...     pol_next, y_next = grid.getPoloidalGrid(i+1)
    ...     pol_last, y_last = grid.getPoloidalGrid(i-1)

    The `getPoloidalGrid()` method ensures that ``y_last <= y <= y_next``

    """

    def __init__(self, poloidal_grids, ycoords, Ly, yperiodic=False, name="fci_grid"):
        try:
            ngrids = len(poloidal_grids)

            # Check this is the same length as ycoords
            assert len(ycoords) == ngrids

            nx = poloidal_grids[0].nx
            nz = poloidal_grids[0].nz
        except TypeError:
            # No len(), assume single poloidal grid
            ngrids = 1
            nx = poloidal_grids.nx
            nz = poloidal_grids.nz

        self.poloidal_grids = poloidal_grids
        self._ngrids = ngrids  # This is an implementation detail, whether we have one or multiple separate grids
        self.ycoords = np.asarray(ycoords)
        self.Ly = Ly
        self.yperiodic = yperiodic

        self._metric_cache = None

        # Define the shape of the grid
        self.shape = (nx, len(ycoords), nz)

    def __repr__(self):
        return "Grid({0}, {1}:{2}:{3}, yperiodic={4})".format(
            self.poloidal_grids,
            np.amin(self.ycoords),
            np.amax(self.ycoords),
            self.ycoords.size,
            self.yperiodic,
        )

    def numberOfPoloidalGrids(self):
        """Returns the number of poloidal grids i.e. number of points in Y

        Returns
        -------
        int
            Number of poloidal grids
        """
        return self.ycoords.size

    def getPoloidalGrid(self, yindex):
        """Returns the poloidal grid and y value at the given y index

        This handles negative values and values out of range, if the
        domain is periodic

        Parameters
        ----------
        yindex : int
            The desired index in y

        Returns
        -------
        :py:obj:`~zoidberg.poloidal_grid.PoloidalGrid`
            The poloidal grid at `yindex`
        float
            The value of the y coordinate at `yindex`

        """
        yindex = int(yindex)

        ny = self.ycoords.size

        if (yindex >= 0) and (yindex < ny):
            # Within index range, so just return
            if self._ngrids == 1:
                # Only one grid
                return self.poloidal_grids, self.ycoords[yindex]
            return self.poloidal_grids[yindex], self.ycoords[yindex]

        # Out of range

        if self.yperiodic:
            # Periodic domain

            # Map index into domain
            y_remap = np.remainder(yindex, ny)  # 0 <= yremap < ny

            # Get number of periods around the domain. Note this can be negative
            nperiods = np.floor(float(yindex) / float(ny))

            ycoord = self.ycoords[y_remap] + nperiods * self.Ly

            if self._ngrids == 1:
                return self.poloidal_grids, ycoord
            return self.poloidal_grids[y_remap], ycoord

        # Not periodic

        if yindex < 0:
            return None, 0.0  # Hit the lower end in Y
        return None, self.Ly  # Hit the upper end in Y

    def metric(self):
        """Return the metric tensor, dx and dz

        Returns
        -------
        dict
            Dictionary containing:
            - **dx, dy, dz**: Grid spacing
            - **gxx, gxz, gyy, gzz**: Covariant components
            - **g_xx, g_xz, g_yy, g_zz**: Contravariant components
        """

        if self._metric_cache:
            return self._metric_cache
        # Gather dx,dz and x-z metrics from poloidal slices
        dx = np.zeros(self.shape)
        dz = np.zeros(self.shape)

        gxx = np.zeros(self.shape)
        gxz = np.zeros(self.shape)
        gzz = np.zeros(self.shape)

        g_xx = np.zeros(self.shape)
        g_xz = np.zeros(self.shape)
        g_zz = np.zeros(self.shape)

        J = np.zeros(self.shape)
        has_J = 0
        # Separate grids for each slice
        for y in range(self.shape[1]):
            pol_metric = self.getPoloidalGrid(y)[0].metric()
            dx[:, y, :] = pol_metric["dx"]
            dz[:, y, :] = pol_metric["dz"]

            gxx[:, y, :] = pol_metric["gxx"]
            gxz[:, y, :] = pol_metric["gxz"]
            gzz[:, y, :] = pol_metric["gzz"]

            g_xx[:, y, :] = pol_metric["g_xx"]
            g_xz[:, y, :] = pol_metric["g_xz"]
            g_zz[:, y, :] = pol_metric["g_zz"]

            if "J" in pol_metric:
                J[:, y, :] = pol_metric["J"]
                has_J += 1
                assert has_J > 0, "J only in some slices!"
            else:
                has_J -= 1
                assert has_J < 0, "J only in some slices!"

        # Calculate the gradient of the y coordinate w.r.t index
        # To avoid edge effects, repeat array three times then take the middle
        ycoords = np.concatenate(
            (self.ycoords - self.Ly, self.ycoords, self.ycoords + self.Ly)
        )
        ny = self.ycoords.size
        if ny == 1:
            dy = np.array([self.Ly])
        else:
            dy = np.gradient(ycoords[ny : (2 * ny)])

        dy3d = np.zeros(self.shape)
        dy3d[...] = dy[None, :, None]

        # Note: These y metrics are for Cartesian coordinates
        # If in cylindrical coordinates then these should be different
        g_yy = np.ones(self.shape)
        gyy = np.ones(self.shape)

        self._metric_cache = {
            "dx": dx,
            "dy": dy3d,
            "dz": dz,
            "gyy": gyy,
            "g_yy": g_yy,
            "gxx": gxx,
            "g_xx": g_xx,
            "gxz": gxz,
            "g_xz": g_xz,
            "gzz": gzz,
            "g_zz": g_zz,
        }

        if has_J > 0:
            self._metric_cache["J"] = J
        return self._metric_cache


def rectangular_grid(
    nx, ny, nz, Lx=1.0, Ly=10.0, Lz=1.0, xcentre=0.0, zcentre=0.0, yperiodic=False
):
    """Create a rectangular grid in (x,y,z)

    Here y is along the magnetic field (typically toroidal angle), and
    (x,z) are in the poloidal plane

    Parameters
    ----------
    nx, ny, nz : int
        Number of points in x, y, z
    Lx, Ly, Lz : float, optional
        Size of the domain in x, y, z
    xcentre, zcentre : float, optional
        The middle of the domain
    yperiodic : bool, optional
        Determines if the y direction is periodic

    Returns
    -------
    `Grid`
        A `Grid` representing a rectangular domain
    """

    # In this simple case we only need one RectangularPoloidalGrid
    # to represent all poloidal planes

    poloidal_grid = RectangularPoloidalGrid(nx, nz, Lx, Lz, xcentre, zcentre)

    if yperiodic:
        ycoords = np.linspace(0.0, Ly, ny, endpoint=False)
    else:
        # Doesn't include the end points
        ycoords = (np.arange(ny) + 0.5) * Ly / float(ny)

    return Grid(poloidal_grid, ycoords, Ly, yperiodic=yperiodic)


if __name__ == "__main__":
    grid = rectangular_grid(10, 10, 10)

    p = grid.getPoloidalGrid(-2)

    print(grid)



from mms_helper import extend, do_tests, plt, bc, np
from parallel_grid import grids


set1 = (
    lambda grid: np.sin(extend(grid["one"] * grid.phi) * 5),
    lambda grid: 5 * np.cos(extend(grid["one"] * grid.phi) * 5),
    "sin(phi)",
)
set2 = (
    lambda grid: np.sin(extend(grid["one"] * grid.phi + grid.r_minor + grid.theta) * 5)
    * extend(np.cos(grid.theta)),
    lambda grid: 5
    * np.cos(extend(grid["one"] * grid.phi + grid.r_minor + grid.theta) * 5)
    * extend(np.cos(grid.theta)),
    "sin(phi*r*theta)*cos(theta)",
)


if __name__ == "__main__":
    bc.init("-d mms -q -q -q")

    do_tests(grids, *set1, bc.DDY)
    do_tests(grids, *set2, bc.DDY)

    plt.show()



from multiprocessing import Pool

import numpy as np
from scipy.integrate import odeint

from .field import _set_config


class FieldTracer(object):
    """A class for following magnetic field lines

    Parameters
    ----------
    field : :py:obj:`~zoidberg.field.MagneticField`
        A Zoidberg MagneticField instance

    """

    def __init__(self, field):
        self.field_direction = field.field_direction

    def follow_field_lines(self, x_values, z_values, y_values, rtol=None):
        """Uses field_direction to follow the magnetic field
        from every grid (x,z) point at toroidal angle y
        through a change in toroidal angle dy

        Parameters
        ----------
        x_values : array_like
            Starting x coordinates
        z_values : array_like
            Starting z coordinates
        y_values : array_like
            y coordinates to follow the field line to. y_values[0] is
            the starting position
        rtol : float, optional
            The relative tolerance to use for the integrator. If None,
            use the default value

        Returns
        -------
        result : numpy.ndarray
            Field line ending coordinates

            The first dimension is y, the last is (x,z). The
            middle dimensions are the same shape as [x|z]:
            [0,...] is the initial position
            [...,0] are the x-values
            [...,1] are the z-values
            If x_values is a scalar and z_values a 1D array, then result
            has the shape [len(y), len(z), 2], and vice-versa.
            If x_values and z_values are 1D arrays, then result has the shape
            [len(y), len(x), 2].
            If x_values and z_values are 2D arrays, then result has the shape
            [len(y), x.shape[0], x.shape[1], 2].

        """

        # Ensure all inputs are NumPy arrays
        x_values = np.atleast_1d(x_values)
        y_values = np.atleast_1d(y_values)
        z_values = np.atleast_1d(z_values)

        if len(y_values) < 2:
            raise ValueError("There must be at least two elements in y_values")
        if len(y_values.shape) > 1:
            raise ValueError("y_values must be 1D")

        if x_values.shape != z_values.shape:
            # Make the scalar the same shape as the array
            if x_values.size == 1:
                x_values = np.zeros(z_values.shape) + x_values
            elif z_values.size == 1:
                z_values = np.zeros(x_values.shape) + z_values
            else:
                raise ValueError(
                    "x_values and z_values must be the same size, or one must be a scalar"
                )

        array_shape = x_values.shape

        # Position vector must be 1D - so flatten before passing to
        # integrator, then reshape after
        if len(x_values.shape) > 1:
            x_values = x_values.flatten()
            z_values = z_values.flatten()

        # If too many points are calculated at once (>~ 1e4) then
        # memory errors occur.

        if len(x_values) < 1000:
            position = np.column_stack((x_values, z_values)).flatten()

            result = odeint(
                self.field_direction, position, y_values, args=(True,), rtol=rtol
            )
        else:
            # Split into smaller pieces

            nchunks = int(len(x_values) / 1000)

            # Splitting x and z values separately to avoid ending up with
            # an odd number of points in position arrays
            x_values = np.array_split(x_values, nchunks)
            z_values = np.array_split(z_values, nchunks)

            # Combine x,z into flattened position arrays
            chunks = [
                np.column_stack((x, z)).flatten() for x, z in zip(x_values, z_values)
            ]

            # Process in chunks. Note: multiprocessing has trouble with closures
            # so fails in Python 2. Python 3 may work
            results = [
                odeint(self.field_direction, chunk, y_values, args=(True,), rtol=rtol)
                for chunk in chunks
            ]

            # Concatenate results into a single array
            result = np.concatenate(results, axis=1)

        return result.reshape(y_values.shape + array_shape + (2,))


class FieldTracerReversible(object):
    """Traces magnetic field lines in a reversible way by using
    trapezoidal integration:

    .. math::

       pos_{n+1} = pos_n + 0.5*( f(pos_n) + f(pos_{n+1}) )*dy

    This requires a Newton iteration to solve the nonlinear set of equations
    for the unknown ``pos_{n+1}``.

    Parameters
    ----------
    field : :py:obj:`~zoidberg.field.MagneticField`
        A Zoidberg MagneticField instance
    rtol : float, optional
        Tolerance applied to changes in dx**2 + dz**2
    eps : float, optional
        Change in x,z used to calculate finite differences of magnetic
        field direction
    nsteps : int, optional
        Number of sub-steps between outputs

    """

    def __init__(self, field, rtol=1e-8, eps=1e-5, nsteps=20):
        self.field_direction = field.field_direction
        self.rtol = float(rtol)
        self.eps = float(eps)
        self.nsteps = int(nsteps)

    def follow_field_lines(
        self, x_values, z_values, y_values, rtol=None, eps=None, nsteps=None
    ):
        """Uses field_direction to follow the magnetic field
        from every grid (x,z) point at toroidal angle y
        through a change in toroidal angle dy

        Parameters
        ----------
        x_values : array_like
            Starting x coordinates
        z_values : array_like
            Starting z coordinates
        y_values : array_like
            y coordinates to follow the field line to. y_values[0] is
            the starting position
        rtol : float, optional
            Tolerance applied to changes in dx**2 + dz**2. If None,
            use the default value
        eps : float, optional
            Change in x,z used to calculate finite differences of magnetic
            field direction
        nsteps : int, optional
            Number of sub-steps between outputs

        Returns
        -------
        result : numpy.ndarray
            Field line ending coordinates

            The first dimension is y, the last is (x,z). The
            middle dimensions are the same shape as [x|z]:
            [0,...] is the initial position
            [...,0] are the x-values
            [...,1] are the z-values
            If x_values is a scalar and z_values a 1D array, then result
            has the shape [len(y), len(z), 2], and vice-versa.
            If x_values and z_values are 1D arrays, then result has the shape
            [len(y), len(x), 2].
            If x_values and z_values are 2D arrays, then result has the shape
            [len(y), x.shape[0], x.shape[1], 2].

        """

        # Check settings, use defaults if not given
        if rtol is None:
            rtol = self.rtol  # Value set in __init__
        rtol = float(rtol)
        if eps is None:
            eps = self.eps
        eps = float(eps)
        if nsteps is None:
            nsteps = self.nsteps
        nsteps = int(nsteps)

        # Ensure all inputs are NumPy arrays
        x_values = np.atleast_1d(x_values)
        y_values = np.atleast_1d(y_values)
        z_values = np.atleast_1d(z_values)

        if len(y_values) < 2:
            raise ValueError("There must be at least two elements in y_values")
        if len(y_values.shape) > 1:
            raise ValueError("y_values must be 1D")

        if x_values.shape != z_values.shape:
            # Make the scalar the same shape as the array
            if x_values.size == 1:
                x_values = np.zeros(z_values.shape) + x_values
            elif z_values.size == 1:
                z_values = np.zeros(x_values.shape) + z_values
            else:
                raise ValueError(
                    "x_values and z_values must be the same size, or one must be a scalar"
                )

        array_shape = x_values.shape

        result = np.zeros((len(y_values),) + array_shape + (2,))

        # Starting position
        x_pos = x_values
        z_pos = z_values
        y_pos = y_values[0]

        result[0, ..., 0] = x_pos
        result[0, ..., 1] = z_pos

        for yindex, y_next in enumerate(y_values[1:]):
            yindex += 1  # Since we chopped off the first y_value

            # Split into sub-steps
            dy = (y_next - y_pos) / float(nsteps)
            for step in range(nsteps):
                # Evaluate gradient at current position
                dxdy, dzdy = self.field_direction((x_pos, z_pos), y_pos)

                # Half-step values
                x_half = x_pos + 0.5 * dxdy * dy
                z_half = z_pos + 0.5 * dzdy * dy

                # Now need to find the roots of a nonlinear equation
                #
                # f = 0.5*(dpos/dy)(pos_next)*dy - (pos_next - pos_half) = 0
                #

                # Use Euler method to get starting guess
                x_pos += dxdy * dy
                z_pos += dzdy * dy
                y_pos += dy

                while True:
                    dxdy, dzdy = self.field_direction((x_pos, z_pos), y_pos)

                    # Calculate derivatives (Jacobian matrix) by finite difference
                    dxdy_xe, dzdy_xe = self.field_direction((x_pos + eps, z_pos), y_pos)
                    dxdy_x = (dxdy_xe - dxdy) / eps
                    dzdy_x = (dzdy_xe - dzdy) / eps

                    dxdy_ze, dzdy_ze = self.field_direction((x_pos, z_pos + eps), y_pos)
                    dxdy_z = (dxdy_ze - dxdy) / eps
                    dzdy_z = (dzdy_ze - dzdy) / eps

                    # The function we are trying to find the roots of:
                    fx = 0.5 * dxdy * dy - x_pos + x_half
                    fz = 0.5 * dzdy * dy - z_pos + z_half

                    # Now have a linear system to solve
                    #
                    # (x_pos)  -= ( dfx/dx   dfx/dz )^-1 (fx)
                    # (z_pos)     ( dfz/dx   dfz/dz )    (fz)

                    dfxdx = 0.5 * dxdy_x * dy - 1.0
                    dfxdz = 0.5 * dxdy_z * dy
                    dfzdx = 0.5 * dzdy_x * dy
                    dfzdz = 0.5 * dzdy_z * dy - 1.0

                    determinant = dfxdx * dfzdz - dfxdz * dfzdx
                    # Note: If determinant is too small then dt should be reduced

                    dx = (dfzdz * fx - dfxdz * fz) / determinant
                    dz = (dfxdx * fz - dfzdx * fx) / determinant

                    x_pos -= dx
                    z_pos -= dz
                    # Check for convergence within tolerance
                    if np.amax(dx**2 + dz**2) < rtol:
                        break
                # Finished Newton iteration, taken step to (x_pos,y_pos,z_pos)
            # Finished sub-steps, reached y_pos = y_next
            result[yindex, ..., 0] = x_pos
            result[yindex, ..., 1] = z_pos

        return result


def trace_poincare(
    magnetic_field, xpos, zpos, yperiod, nplot=3, y_slices=None, revs=20, nover=20
):
    """Trace a Poincare graph of the field lines

    Does no plotting, see :py:func:`zoidberg.plot.plot_poincare`

    Parameters
    ----------
    magnetic_field : :py:obj:`~zoidberg.field.MagneticField`
        Magnetic field object
    xpos, zpos : array_like
        Starting X, Z locations
    yperiod : float
        Length of period in y domain
    nplot : int, optional
        Number of equally spaced y-slices to trace to
    y_slices : list of ints
        List of y-slices to plot; overrides `nplot`
    revs : int, optional
        Number of revolutions (times around y)
    nover : int, optional
        Over-sample. Produced additional points in y then discards.
        This seems to be needed for accurate results in some cases

    Returns
    -------
    coords, y_slices
        coords is a Numpy array of data::

            [revs, nplot, ..., R/Z]

        where the first index is the revolution, second is the y slice,
        and last is 0 for R, 1 for Z. The middle indices are the shape
        of the input xpos,zpos

    """

    if nplot is None and y_slices is None:
        raise ValueError("nplot and y_slices cannot both be None")

    if y_slices is not None:
        y_slices = np.asarray(y_slices)

        if np.amin(y_slices) < 0.0 or np.amax(y_slices) > yperiod:
            raise ValueError(
                "y_slices must all be between 0.0 and yperiod ({yperiod})".format(
                    yperiod=yperiod
                )
            )
        # Make sure y_slices is monotonically increasing
        y_slices.sort()
        # If y_slices is given, then nplot is the number of slices
        nplot = len(y_slices)
    else:
        # nplot equally spaced phi slices
        nplot = int(nplot)
        y_slices = np.linspace(0, yperiod, nplot, endpoint=False)

    # Extend the domain from [0,yperiod] to [0,revs*yperiod]

    revs = int(revs)
    y_values = y_slices[:]
    for n in np.arange(1, revs):
        y_values = np.append(y_values, n * yperiod + y_values[:nplot])

    nover = int(nover)  # Over-sample
    y_values_over = np.zeros((nplot * revs * nover - (nover - 1)))
    y_values_over[::nover] = y_values
    for i in range(1, nover):
        y_values_over[i::nover] = (float(i) / float(nover)) * y_values[1:] + (
            float(nover - i) / float(nover)
        ) * y_values[:-1]

    # Starting location
    xpos = np.asarray(xpos)
    zpos = np.asarray(zpos)

    field_tracer = FieldTracer(magnetic_field)
    result = field_tracer.follow_field_lines(xpos, zpos, y_values_over)

    result = result[::nover, ...]  # Remove unneeded points

    # Reshape data. Loops fastest over planes (nplot)
    result = np.reshape(result, (revs, nplot) + result.shape[1:])

    return result, y_slices


def _get_value(*args):
    for k in args:
        if k is not None:
            return k
    raise ValueError(f"Expected at least one non-None value in {args}")


class FieldTracerWeb:
    """A class for following magnetic field lines

    Parameters
    ----------
    field : :py:obj:`~zoidberg.field.MagneticField`
        A Zoidberg MagneticField instance

    """

    def __init__(
        self,
        config=None,
        configId=None,
        timeout=0.1,
        chunk=10000,
        stepsize=None,
        retry=None,
        run_timeout=None,
    ):
        try:
            import requests
        except ModuleNotFoundError:
            requests = None
        # First check whether the webservice is available, as the default timeout is rather slow.
        self.url = "http://esb.ipp-hgw.mpg.de:8280/services/FieldLineProxy?wsdl"
        if requests:
            requests.get(self.url, timeout=timeout).raise_for_status()

        self.timeout = run_timeout
        self.retry = retry
        self.config = config
        self.configId = configId
        self.stepsize = stepsize
        if self.config:
            assert self.configId is None
        else:
            assert self.configId is not None

        # Check tracing direction
        pos = self.flt.types.Points3D()
        pos.x1 = [6]
        pos.x2 = [0]
        pos.x3 = [0]

        lineTask = self.flt.types.LineTracing()
        lineTask.numSteps = 1

        task = self.flt.types.Task()
        task.step = 0.001
        task.lines = lineTask

        line = self.flt.service.trace(pos, self.getConfig(), task, None, None).lines[0]
        phi = np.arctan2(line.vertices.x2, line.vertices.x1)[-1]
        assert phi != 0
        self.isforward = phi > 0
        self.chunk = chunk

    def follow_field_lines(
        self,
        x_values,
        z_values,
        y_values,
        rtol=None,
        stepsize=None,
        timeout=None,
        retry=None,
        chunk=None,
    ):
        """Uses field_direction to follow the magnetic field
        from every grid (x,z) point at toroidal angle y
        through a change in toroidal angle dy

        Parameters
        ----------
        x_values : array_like
            Starting x coordinates
        z_values : array_like
            Starting z coordinates
        y_values : array_like
            y coordinates to follow the field line to. y_values[0] is
            the starting position
        rtol : float, optional
            Currently ignored. Use stepsize instead
        stepsize : float, optional
            The stepsize for the integrator. Sets accuracy and speed.
        timeout : float, optional
            How long to wait. Sometimes the server does not return and
            the client keeps waiting for ever, otherwise.
        retry : int, optional
            How often to retry failed calculations
        chunk: int or None, optional
            Set chunking size, overwrite default if not None

        Returns
        -------
        result : numpy.ndarray
            Field line ending coordinates

            The first dimension is y, the last is (x,z). The
            middle dimensions are the same shape as [x|z]:
            [0,...] is the initial position
            [...,0] are the x-values
            [...,1] are the z-values
            If x_values is a scalar and z_values a 1D array, then result
            has the shape [len(y), len(z), 2], and vice-versa.
            If x_values and z_values are 1D arrays, then result has the shape
            [len(y), len(x), 2].
            If x_values and z_values are 2D arrays, then result has the shape
            [len(y), x.shape[0], x.shape[1], 2].

        """
        retry = _get_value(retry, self.retry, 3)
        timeout = _get_value(timeout, self.timeout, 1800)

        # Ensure all inputs are NumPy arrays
        x_values = np.atleast_1d(x_values)
        y_values = np.atleast_1d(y_values)
        z_values = np.atleast_1d(z_values)

        if len(y_values) < 2:
            raise ValueError("There must be at least two elements in y_values")
        if len(y_values.shape) > 1:
            raise ValueError("y_values must be 1D")

        if x_values.shape != z_values.shape:
            # Make the scalar the same shape as the array
            if x_values.size == 1:
                x_values = np.zeros(z_values.shape) + x_values
            elif z_values.size == 1:
                z_values = np.zeros(x_values.shape) + z_values
            else:
                raise ValueError(
                    "x_values and z_values must be the same size, or one must be a scalar"
                )

        array_shape = x_values.shape

        # Position vector must be 1D - so flatten before passing to
        # integrator, then reshape after
        if len(x_values.shape) > 1:
            x_values = x_values.flatten()
            z_values = z_values.flatten()
        chunk = chunk or self.chunk
        if self.config is None and x_values.size > chunk:
            with Pool(12) as pool:

                def start(i):
                    return pool.apply_async(
                        self._follow_field_lines,
                        (
                            x_values[i : i + chunk],
                            z_values[i : i + chunk],
                            y_values,
                            rtol,
                            stepsize,
                        ),
                    )

                # Start parallel calculation
                results = [start(i) for i in range(0, len(x_values), chunk)]
                # Wait for result and combine
                for j, i in enumerate(range(0, len(x_values), chunk)):
                    while retry > 0:
                        try:
                            results[j] = results[j].get(timeout=timeout)
                            break
                        except Exception as e:
                            retry -= 1
                            if not retry > 0:
                                raise
                            print(
                                f"Fieldlinetracer failed {e} - going to retry {retry} times"
                            )
                            results[j] = start(i)
                try:
                    results = np.concatenate(results, axis=1)
                except ValueError:
                    print(results)
                    raise
                    raise ValueError(
                        f"np.concatenate failed, shapes where {', '.join([str(x.shape) for x in results])}."
                    )
        else:
            results = [
                self._follow_field_lines(
                    x_values[i : i + chunk],
                    z_values[i : i + chunk],
                    y_values,
                    rtol,
                    stepsize,
                )
                for i in range(0, len(x_values), chunk)
            ]
            # Wait for result and combine
            results = np.concatenate(results, axis=1)

        return results.reshape(y_values.shape + array_shape + (2,))

    def getConfig(self):
        if self.config is not None:
            self.config.inverseField = False
            return self.config
        config = self.flt.types.MagneticConfig()
        config = _set_config(config, self.configId)
        config.inverseField = False
        return config

    @property
    def flt(self):
        from osa import Client

        return Client(self.url)

    def _follow_field_lines(self, x_values, z_values, y_values, rtol, stepsize):
        from time import sleep

        sleep(np.random.random(1)[0])
        p = self.flt.types.Points3D()
        p.x1 = x_values * np.cos(y_values[0])
        p.x2 = x_values * np.sin(y_values[0])
        p.x3 = z_values

        task = self.flt.types.Task()
        task.step = stepsize or self.stepsize or 0.01
        task.linesPhi = self.flt.types.LinePhiSpan()

        def xor(x, y):
            return bool((x and not y) or (not x and y))

        dphi = y_values[1] - y_values[0]
        config = self.getConfig()
        config.inverseField = xor(dphi > 0, self.isforward)
        result = np.empty((len(y_values), len(p.x1), 2))
        result[0, :, 0] = x_values
        result[0, :, 1] = z_values
        phi0 = y_values[0]
        for i, phi in enumerate(y_values):
            if i == 0:
                continue
            task.linesPhi.phi = phi - phi0
            res = self.flt.service.trace(p, config, task)
            for j, curve in enumerate(res.lines):
                ps = curve.vertices
                result[i, j, 0] = np.sqrt(ps.x2[-1] ** 2 + ps.x1[-1] ** 2)
                result[i, j, 1] = ps.x3[-1]
        return result



import matplotlib.pyplot as plt
import xarray as xr
import sys


def plot(ds, var, i):
    print(ds[var].min().data, ds[var].max().data)
    plt.pcolormesh(ds.R[:, i], ds.Z[:, i], ds[var][:, i])
    plt.colorbar()
    plt.show()


with xr.open_dataset(sys.argv[1]) as ds:
    plot(ds, sys.argv[2], int(sys.argv[3]))



import numpy as np

from .field import CurvedSlab, Slab


def test_slab():
    # Slab with no shear (Bzprime = 0)
    slab = Slab(By=0.5, Bz=0.2, Bzprime=0.0)

    # Create an array of coordinates
    x = np.arange(10)
    z = x

    Bx = slab.Bxfunc(x, z, 0.0)
    assert Bx.shape == x.shape
    assert np.allclose(Bx, 0.0)

    By = slab.Byfunc(x, z, 0.0)
    assert By.shape == x.shape
    assert np.allclose(By, 0.5)

    Bz = slab.Bzfunc(x, z, 0.0)
    assert Bz.shape == x.shape
    assert np.allclose(Bz, 0.2)


def test_curved_slab():
    # Slab with no shear (Bzprime = 0)
    slab = CurvedSlab(By=0.5, Bz=0.2, Bzprime=0.0)

    # Create an array of coordinates
    x = np.arange(10)
    z = x

    Bx = slab.Bxfunc(x, z, 0.0)
    assert Bx.shape == x.shape
    assert np.allclose(Bx, 0.0)

    By = slab.Byfunc(x, z, 0.0)
    assert By.shape == x.shape
    assert np.allclose(By, 0.5)

    Bz = slab.Bzfunc(x, z, 0.0)
    assert Bz.shape == x.shape
    assert np.allclose(Bz, 0.2)



import numpy as np

from .smooth import find_outliers, smooth


def test_square():
    ### test smooth function on a square
    ### create square
    square_r = np.zeros(400)
    square_z = np.zeros(400)
    square_r[0:100] = np.linspace(0, 1, num=100)
    square_z[0:100] = np.linspace(0, 0, num=100)
    square_r[100:200] = np.linspace(1, 1, num=100)
    square_z[100:200] = np.linspace(0, 1, num=100)
    square_r[200:300] = np.linspace(1, 0, num=100)
    square_z[200:300] = np.linspace(1, 1, num=100)
    square_r[300:400] = np.linspace(0, 0, num=100)
    square_z[300:400] = np.linspace(1, 0, num=100)
    ### smooth square
    newr, newz = smooth(square_r, square_z, cutoff=0.75, bounds=10)
    outliers = find_outliers(newr, newz, 0.75)
    assert len(outliers) == 0


def run_npolygons(n=10, m=1000):
    ### test smooth function on a symmetric n-polygone
    ### create n-polygone
    circle = np.linspace(0, 2 * np.pi, num=n, endpoint=False)
    ends_r = np.cos(circle)
    ends_z = np.sin(circle)
    phi_r = np.linspace(0, 2 * np.pi, n, endpoint=False)
    phi_z = np.linspace(0, 2 * np.pi, m, endpoint=False)
    test_r = np.interp(phi_z, phi_r, ends_r, period=np.pi * 2)
    test_z = np.interp(phi_z, phi_r, ends_z, period=np.pi * 2)
    ### smooth n-polygone
    newr, newz = smooth(test_r, test_z, 0.96, 10)
    outliers = find_outliers(newr, newz, 0.96, 30)
    return outliers, test_r, test_z, newr, newz


def test_assert_npolygons(n=10, m=100):
    ### test the smooth function and evaluates the outcome
    ### finds nothing after smoothing
    outliers, testr, testz, newr, newz = run_npolygons(n=10, m=1000)
    assert not outliers and testz.all() != newz.all()
    ### no smoothing done
    outliers, testr, testz, newr, newz = run_npolygons(n=1000, m=10000)
    assert not outliers and testz.all() == newz.all()
    ### find outliers after smoothing
    outliers, testr, testz, newr, newz = run_npolygons(n=5, m=1000)
    assert outliers and testz.all() != newz.all()


if __name__ == "__main__":
    test_assert_npolygons(n=10, m=1000)



from itertools import chain, product

import numpy as np

from . import field, grid, zoidberg


def test_make_maps_slab():
    nx = 5
    ny = 8
    nz = 7

    # Create a straight magnetic field in a slab
    straight_field = field.Slab(By=1.0, Bz=0.0, Bzprime=0.0)

    # Create a rectangular grid in (x,y,z)
    rectangle = grid.rectangular_grid(nx, ny, nz)

    # Two parallel slices in each direction
    nslice = 2

    # Calculate forwards and backwards maps
    maps = zoidberg.make_maps(rectangle, straight_field, nslice=nslice)

    # Since this is a straight magnetic field in a simple rectangle,
    # all the maps should be the same, and should be the identity
    identity_map_x, identity_map_z = np.meshgrid(
        np.arange(nx), np.arange(nz), indexing="ij"
    )

    # Check that maps has the required forward and backward index variables
    offsets = chain(range(1, nslice + 1), range(-1, -(nslice + 1), -1))
    field_line_maps = ["xt_prime", "zt_prime"]

    for field_line_map, offset in product(field_line_maps, offsets):
        var = zoidberg.parallel_slice_field_name(field_line_map, offset)
        print("Current field: ", var)
        assert var in maps

        # Each map should have the same shape as the grid
        assert maps[var].shape == (nx, ny, nz)

        # The first/last abs(offset) points are not valid, so ignore those
        interior_range = (
            range(ny - abs(offset)) if offset > 0 else range(abs(offset), ny)
        )
        # Those invalid points should be set to -1
        end_slice = slice(-1, -(offset + 1), -1) if offset > 0 else slice(0, -offset)
        identity_map = identity_map_x if "x" in var else identity_map_z

        for y in interior_range:
            assert np.allclose(maps[var][:, y, :], identity_map)

        # The end slice should hit a boundary
        assert np.allclose(maps[var][:, end_slice, :], -1.0)


def test_make_maps_straight_stellarator():
    nx = 5
    ny = 6
    nz = 7

    # Create magnetic field
    magnetic_field = field.StraightStellarator(radius=np.sqrt(2.0))

    # Create a rectangular grid in (x,y,z)
    rectangle = grid.rectangular_grid(
        nx, ny, nz, Lx=1.0, Lz=1.0, Ly=10.0, yperiodic=True
    )

    # Here both the field and and grid are centred at (x,z) = (0,0)
    # and the rectangular grid here fits entirely within the coils

    maps = zoidberg.make_maps(rectangle, magnetic_field)



import boutpp as bc
import matplotlib.pyplot as plt
import numpy as np

from mms_helper import extend, do_tests
from poloidal_grid import grids


def inp(grid):
    t = extend(grid.theta)
    return np.sin(t)


def ana(grid):
    t = extend(grid.theta)
    return np.cos(t)


if __name__ == "__main__":
    bc.init("-d mms -q -q -q")

    do_tests(grids, inp, ana, "sin(theta)", bc.DDZ)

    plt.show()



#!/usr/bin/env python

# Demonstrates how to generate 3D plots of magnetic field lines

import zoidberg

# Size of the domain in y (periodic)
yperiod = 10.0

# Define magnetic field
magnetic_field = zoidberg.field.StraightStellarator(
    I_coil=0.4, radius=1.0, yperiod=yperiod
)

# Make 3D plot
zoidberg.plot.plot_3d_field_line(
    magnetic_field,
    0.3,  # x starting position
    0.0,  # z starting position
    yperiod,  # Periodicity of y domain
    cycles=20,
)



# -*- coding: utf-8 -*-

# Adapted from http://stackoverflow.com/a/15860757/2043465

import time
from sys import stdout


def update_progress(progress, barLength=10, ascii=False, **kwargs):
    """Displays or updates a console progress bar

    Accepts a float between 0 and 1. Any int will be converted to a float.
    A value under 0 represents a 'halt'.
    A value at 1 or bigger represents 100%

    Parameters
    ----------
    progress : float
        Number between 0 and 1
    barLength : int, optional
        Length of the progress bar
    ascii : bool, optional
        If True, use '#' as the progress indicator, otherwise use a
        Unicode character (the default)

    """

    if ascii:
        cursor = "#"
    else:
        cursor = "█"

    status = ""
    if isinstance(progress, int):
        progress = float(progress)
    if not isinstance(progress, float):
        progress = 0
        status = "error: progress var must be float\r\n"
    if progress < 0:
        progress = 0
        status = "Halt...\r\n"
    if progress >= 1:
        progress = 1
        status = "Done...\r\n"
    block = int(round(barLength * progress))
    text = "\rPercent: [{prog:-<{len}}] {perc:6.2f}% {stat}".format(
        len=barLength, prog=cursor * block, perc=progress * 100, stat=status
    )

    # Secret undocumented feature
    if "zoidberg" in kwargs:
        if kwargs["zoidberg"]:
            if barLength < 40:
                barLength = 40
            if ascii:
                face = " (;,,,;) "
                ink = "#"
            else:
                face = " (°,,,°) "
                ink = "█"

            open_claw = "(\\/)"
            closed_claw = "(|)"

            if int(progress * barLength) % 2:
                left_claw = open_claw
                right_claw = closed_claw
            else:
                left_claw = closed_claw
                right_claw = open_claw

            zb = left_claw + face + right_claw
            zb_middle = int(len(zb) / 2)
            start = int(round((barLength - zb_middle) * progress))
            text = "\rProgress: [{start}{zb}{rest}] {perc:6.2f}% {stat}".format(
                start=ink * start,
                zb=zb,
                perc=progress * 100,
                rest="-" * (barLength - start - zb_middle),
                stat=status,
            )

    stdout.write(text)
    stdout.flush()


def format_time(secs):
    ret = ""
    out = False
    for name, val in [("m", 60), ("h", 3600), ("d", 60 * 60 * 24)][::-1]:
        if secs >= val:
            out = True
        if out:
            cur = int(secs // val)
            ret += f"{cur:02d}{name}"
            secs -= cur * val
    ret += f"{secs:02.0f}s"
    return ret


class Progress(object):
    def __init__(self):
        pass

    def update(self, prog):
        needed = time.time() - self.t0
        print(
            f"{prog * 100:6.2f} % ... {format_time(needed)}>{format_time(needed / (prog) - needed) if prog else 0}",
            end="\r",
        )

    def __enter__(self):
        self.t0 = time.time()
        return self

    def __exit__(self, *args):
        self.update(1)
        print()



import warnings

try:
    import matplotlib.pyplot as plt
except ImportError:
    plt = None
import numpy as np
from scipy.optimize import least_squares

from .rzline import RZline


def _fft(args, freq):
    out = args[0]
    for i0, (s, c) in enumerate(zip(args[1::2], args[2::2])):
        i = i0 + 1
        out += np.sin(i * freq) * s
        out += np.cos(i * freq) * c
    return out


def _err(args, data, freq):
    freq = np.linspace(-np.pi, np.pi, len(data), endpoint=False)
    return _fft(args, freq) - data


class SpectralLine:
    """
    similar to RZline, but based on fourier harmonics
    """

    def __init__(self, line, num):
        """
        Create an SpectralLine from a `line` using `num` modes.

        line: RZline
        num: integer
        """
        freq = line.distance()
        freq = freq[:-1] * np.pi * 2 / freq[-1]
        n = 2 * num + 1
        self.resR = least_squares(_err, np.zeros(n), args=(line.R, freq))
        self.resZ = least_squares(_err, np.zeros(n), args=(line.Z, freq))

        self.cost = self.resR.cost + self.resZ.cost
        self.orgLine = line

    def Rvalue(self, theta):
        return _fft(self.resR.x, theta)

    def Zvalue(self, theta):
        return _fft(self.resZ.x, theta)

    def equallySpaced(self, num=None):
        freq = np.linspace(0, np.pi * 2, num or len(self.orgLine.R))
        return RZline(self.Rvalue(freq), self.Zvalue(freq))

    def plot(self, axis=None, show=True):
        """Plot the RZline, either on the given axis or a new figure

        Parameters
        ----------
        axis : matplotlib axis, optional
            A matplotlib axis to plot on. By default a new figure
            is created
        show : bool, optional
            Calls plt.show() at the end

        Returns
        -------
        axis
            The matplotlib axis that was used

        """
        if plt is None:
            warnings.warn("matplotlib not available, unable to plot")
            return None

        if axis is None:
            fig = plt.figure()
            axis = fig.add_subplot(1, 1, 1)

        theta = np.linspace(0, 2 * np.pi, 100, endpoint=True)
        axis.plot(self.Rvalue(theta), self.Zvalue(theta), "k-")
        axis.plot(self.orgLine.R, self.orgLine.Z, "ro")

        if show:
            plt.show()

        return axis



import zoidberg as zb
import numpy as np
import xarray as xr
import os

lst = [16, 32]

print(zb.__path__)
modes = [
    ("const", np.linspace),
    # ("exp", np.geomspace),
]


class FixedCurvedSlab(zb.field.CurvedSlab):
    def Rfunc(self, x, z, phi):
        return x


def gen_name(*args):
    nx, ny, nz, R0, r0, r1, mode = args
    return f"delp_{modes[mode][0]}_{nx}_{ny}_{nz}_{R0}.fci.nc"


def gen_grid(nx, ny, nz, R0, r0, r1, mode=0):
    mode = modes[mode]
    one = np.ones((nx, ny, nz))
    r = mode[1](r0, r1, nx)[:, None]
    theta = np.linspace(0, 2 * np.pi, nz, False)[None, :]
    phi = np.linspace(0, 2 * np.pi / 5, ny, False)
    R = R0 + np.cos(theta) * r
    Z = np.sin(theta) * r
    pol_grid = zb.poloidal_grid.StructuredPoloidalGrid(R, Z)

    field = FixedCurvedSlab(Bz=0, Bzprime=0, Rmaj=R0)
    grid = zb.grid.Grid(pol_grid, phi, 5, yperiodic=True)

    fn = gen_name(*args)

    maps = zb.make_maps(grid, field, quiet=True, MXG=1)
    zb.write_maps(
        grid,
        field,
        maps,
        "tmp.nc",
        metric2d=False,
    )
    with xr.open_dataset("tmp.nc") as ds:
        for k in "g11,g33,g_11,g_33,nx,ny,nz".split(","):
            assert np.all(ds[k] > 0)
        dims = ds.dz.dims
        ds["r_minor"] = dims, one * r[:, None, :]
        ds["phi"] = "y", phi
        ds["theta"] = dims, one * theta[:, None, :]
        ds["one"] = dims, one
        ds.to_netcdf(fn)


def _togen(*args):
    return gen_name(*args), args


grids = {}
for mode in range(len(modes)):
    grids[modes[mode][0]] = [_togen(nz, 2, nz, 1, 0.1, 0.5, mode) + (nz,) for nz in lst]
    # grids[modes[mode][0] + "2"] = [
    #     _togen(nz, 2, 4, 2, 0.1, 1.1, mode) + (nz,) for nz in lst
    # ]
print(grids)

if __name__ == "__main__":
    for todos in grids.values():
        for fn, args, _ in todos:
            gen_grid(*args)
